import json
import os
import pickle
from collections import OrderedDict
from enum import Enum

import mlflow
import yaml
from mlflow.tracking import MlflowClient
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger


def add_to_store_object(source, obj, store=True):
    from pypads.app.pypads import get_current_pads
    pads = get_current_pads()
    if not pads.cache.run_exists(id(source)):
        pads.cache.run_add(id(source), OrderedDict())

    objects_store: OrderedDict = pads.cache.run_get(id(source))

    if id(obj) not in objects_store:
        objects_store[id(obj)] = (obj, store)
    else:
        logger.warning("Object already added to the store")


def get_temp_folder(run=None):
    """
    Get the base folder to log tmp files to. For now it can't be changed. Todo make configurable
    :return:
    """
    from pypads.app.pypads import get_current_pads
    pads = get_current_pads()
    run = run if run else pads.api.active_run()
    if run is None:
        raise ValueError("No active run is defined.")
    return os.path.join(pads.folder, "tmp", run.info.experiment_id, run.info.run_id) + os.path.sep


def get_run_folder():
    """
    Get the folder holding the run information.
    :return:
    """
    run = mlflow.active_run()
    if run is None:
        raise ValueError("No active run is defined.")
    # TODO use artifact download if needed or load artifact.
    return os.path.join(mlflow.get_tracking_uri(), run.info.experiment_id, run.info.run_id)


class WriteFormats(Enum):
    pickle = 'pickle'
    text = 'text'
    yaml = 'yaml'
    json = 'json'


class ReadFormats(Enum):
    pickle = 'pickle'
    txt = 'txt'
    yaml = 'yaml'
    json = 'json'


# extract all tags of runs by experiment id
def all_tags(experiment_id):
    client = MlflowClient(mlflow.get_tracking_uri())
    ds_infos = client.list_run_infos(experiment_id)
    for i in ds_infos:
        yield mlflow.get_run(i.run_id).data.tags


def try_read_artifact(file_name, folder_lookup=True):
    """
    Function to read an artifact from disk
    :param file_name:
    :return:
    """
    # TODO make defensive
    path = file_name
    if folder_lookup:
        base_path = get_run_folder()
        path = os.path.join(base_path, "artifacts", file_name)

    # Functions for the options to read
    def read_text(p):
        with open(p, "r") as fd:
            return fd.read()

    def read_pickle(p):
        try:
            with open(p, "rb") as fd:
                return pickle.load(fd)
        except Exception as e:
            logger.warning("Couldn't read pickle file. " + str(e))

    def read_yaml(p):
        try:
            with open(p, "r") as fd:
                return yaml.full_load(fd)
        except Exception as e:
            logger.warning("Couldn't read artifact as yaml. Trying to read it as text instead. " + str(e))
            return read_text(p)

    def read_json(p):
        try:
            with open(p, "r") as fd:
                return json.load(fd)
        except Exception as e:
            logger.warning("Couldn't read artifact as json. Trying to read it as text instead. " + str(e))
            return read_text(p)

    options = {
        ReadFormats.pickle: read_pickle,
        ReadFormats.txt: read_text,
        ReadFormats.yaml: read_yaml,
        ReadFormats.json: read_json
    }

    read_format = path.split('.')[-1]
    if ReadFormats[read_format]:
        read_format = ReadFormats[read_format]
    else:
        logger.warning("Configured read format " + read_format + " not supported! ")
        return
    try:
        data = options[read_format](path)
    except Exception as e:
        logger.warning("Reading artifact failed for '" + path + "'. " + str(e))
        data = "Cannot view content"
    return data


def try_write_artifact(file_name, obj, write_format, preserve_folder=True):
    """
    Function to write an artifact to disk.
    :param preserve_folder:
    :param write_format:
    :param file_name:
    :param obj:
    :return:
    """
    base_path = get_temp_folder()
    path = os.path.join(base_path, file_name)

    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    # Functions for the options to write to
    def write_text(p, o):
        with open(p + ".txt", "w+") as fd:
            fd.write(str(o))
            return fd.name

    def write_pickle(p, o):
        try:
            with open(p + ".pickle", "wb+") as fd:
                pickle.dump(o, fd)
                return fd.name
        except Exception as e:
            logger.warning("Couldn't pickle output. Trying to save toString instead. " + str(e))
            return write_text(p, o)

    def write_yaml(p, o):
        try:
            with open(p + ".yaml", "w+") as fd:
                if isinstance(o, str):
                    fd.write(o)
                    # TODO check if valid json?
                else:
                    yaml.dump(o, fd)
                return fd.name
        except Exception as e:
            logger.warning("Couldn't write meta as yaml. Trying to save it as json instead. " + str(e))
            return write_json(p, o)

    def write_json(p, o):
        try:
            with open(p + ".json", "w+") as fd:
                if isinstance(o, str):
                    fd.write(o)
                    # TODO check if valid json?
                else:
                    json.dump(o, fd)
                return fd.name
        except Exception as e:
            logger.warning("Couldn't write meta as json. Trying to save it as text instead. " + str(e))
            return write_text(p, o)

    # Options to write to
    options = {
        WriteFormats.pickle: write_pickle,
        WriteFormats.text: write_text,
        WriteFormats.yaml: write_yaml,
        WriteFormats.json: write_json
    }

    # Write to disk
    if isinstance(write_format, str):
        if WriteFormats[write_format]:
            write_format = WriteFormats[write_format]
        else:
            logger.warning("Configured write format " + write_format + " not supported! ")
            return

    path = options[write_format](path, obj)
    if preserve_folder:
        in_folder = os.path.join(base_path, file_name.split(os.sep)[0])
        # Log artifact to mlflow
        if os.path.isdir(in_folder):
            try_mlflow_log(mlflow.log_artifact, in_folder)
        else:
            try_mlflow_log(mlflow.log_artifact, path)
    else:
        try_mlflow_log(mlflow.log_artifact, path)


def _to_artifact_meta_name(name):
    return name + ".artifact"


def _to_metric_meta_name(name):
    return name + ".metric"


def _to_param_meta_name(name):
    return name + ".param"



import json
import os
from glob import glob

from pypads.utils.logging_util import try_read_artifact
from pypads.utils.util import set_in_dict


def find_directory(name, path):
    """
    Searches for a directory with the name starting from the path
    :param name:
    :param path:
    :return: A list of directory paths
    """
    found_paths = []
    for root, dirs, files in os.walk(path):
        if name in dirs:
            found_paths.append(os.path.join(root, name))

    return found_paths


def read_metric_file_contents(path):
    """
    Reads a ASCII text file and returns it as a list of tuples
    :param path: Path for the file
    :return: A list of tuples containing the timestamp, metric value and step number
    """
    result_tuple = []
    with open(path, 'r') as fp:
        content = fp.readlines()

    for line in content:
        result_tuple.append(tuple(line.strip().split(sep=' ')))

    return result_tuple


def get_metrics(root_path):
    """
    Reads all the files within the directories named metric
    :param root_path: Root path from where the search begins
    :return: Dictionary with the name of the metric as key and a list of tuples with timestamp, metric value and step
    """

    metrics_dict = dict()
    metric_names = []
    file_paths = []
    paths = find_directory('metrics', root_path)
    for path in paths:
        file_names = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
        metric_names = metric_names + file_names
        file_paths = file_paths + [os.path.join(path, f) for f in file_names]

    for idx, file in enumerate(file_paths):
        result = read_metric_file_contents(file)
        metrics_dict[metric_names[idx]] = result

    return metrics_dict


def get_params(root_path):
    """
    Finds all the parameters with the folder named params
    :param root_path: Root path from where the search begins
    :return: A nested dictionary with the key as the estimators and the second level containing individual parameters
    """

    params_dict = dict()
    param_names = []
    file_paths = []

    # Find all subdirectories having the name param starting from the root path
    paths = find_directory('params', root_path)
    for path in paths:
        file_names = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]

        # Add that to the file names
        param_names = param_names + file_names
        # Turn file names into absolute paths
        file_paths = file_paths + [os.path.join(path, f) for f in file_names]

    for idx, file_path in enumerate(file_paths):
        param_dict = dict()
        with open(file_path, "r") as fp:
            content = fp.readline()
        content = content.strip()
        param_name = param_names[idx]
        param = param_name.split(sep='.')[-1]
        estimator = param_name[:param_name.rfind('.')]
        param_dict[param] = content
        existing_param_dict = params_dict.get(estimator, dict())
        params_dict[estimator] = {**existing_param_dict, **param_dict}

    return params_dict


def get_tags(root_path):
    """
    Finds all the tags for an experimental run
    :param path:
    :return:
    """
    tags_dict = dict()
    tag_names = []
    file_paths = []
    # Find all subdirectories having the name param starting from the root path
    paths = find_directory('tags', root_path)
    for path in paths:
        file_names = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]

        # Add that to the file names
        tag_names = tag_names + file_names
        # Turn file names into absolute paths
        file_paths = file_paths + [os.path.join(path, f) for f in file_names]

    for idx, file_path in enumerate(file_paths):
        with open(file_path, "r") as fp:
            content = fp.readline()
        content = content.strip()
        tags_dict[tag_names[idx]] = content

    return tags_dict


def consolidate_run_output_files(root_path):
    """
    This function consolidates all the written JSON and text files for an experimental run
    :param root_path: The path of the run from which the search begins
    :return:
    """

    def read_file(path):
        with open(path, 'r') as fp:
            data = json.load(fp)
        return data

    consolidated_dict = dict()

    # Find all the JSON files that will be concantenated by path
    json_files = [y for x in os.walk(root_path) for y in glob(os.path.join(x[0], '*.json'))]
    for json_file_path in json_files:
        consolidated_dict[json_file_path] = read_file(json_file_path)

    # Add the metrics
    consolidated_dict['metrics'] = get_metrics(root_path=root_path)

    # Add the parameters of the experiment
    consolidated_dict['parameters'] = get_params(root_path=root_path)

    # Add the tags of the experiment
    consolidated_dict['tags'] = get_tags(root_path=root_path)

    # Write the result
    output_path = os.path.join(root_path, 'consolidated.json')
    with open(output_path, "w") as fp:
        fp.write(json.dumps(consolidated_dict))


def get_artifacts(path, search=""):
    """
    Get all logged artifacts under path.
    :param path:
    :return:
    """
    #TODO
    results = {"artifacts": dict()}
    idx = len(path.split("/")) - 1
    for root, dirs, files in os.walk(path):
        keys = root.split('/')[idx:]
        if search!="":
            if search == keys[-1]:
                r = {f: try_read_artifact(os.path.join(root,f), folder_lookup=False) for f in files}
                set_in_dict(results, [keys[0], ".".join(keys[1:])], r)
            else:
                continue
        else:
            r = dict.fromkeys(dirs, dict())
            set_in_dict(results, keys, r)
            if files != []:
                r = {f: try_read_artifact(os.path.join(root,f), folder_lookup=False) for f in files}
                set_in_dict(results, keys, r)
    return results



import os
import sys
from abc import abstractmethod
from typing import Union

import mlflow
from mlflow.tracking import MlflowClient
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger
from pypads.app.injections.base_logger import TrackedObject, LoggerOutput
from pypads.model.models import ArtifactMetaModel, MetricMetaModel, ParameterMetaModel, TagMetaModel, MetadataModel
from pypads.utils.logging_util import try_write_artifact, WriteFormats
from pypads.utils.util import string_to_int


class BackendInterface:

    def __init__(self, uri, pypads):
        self._uri = uri
        self._pypads = pypads
        self._managed_result_git = None

        manage_results = self._uri.startswith("git://")

        # If the results should be git managed
        if manage_results:
            result_path = self._uri[5:]
            self._uri = os.path.join(self._uri[5:], "r_" + str(string_to_int(uri)), "experiments")
            self.manage_results(result_path)
            pypads.cache.add('uri', self._uri)

    @abstractmethod
    def manage_results(self, result_path):
        """
        If we should push results for the user use a managed git.
        :param result_path: Path where the result git should be.
        :return:
        """
        raise NotImplementedError("")

    @abstractmethod
    def add_result_remote(self, remote, uri):
        """
        Add a remote to track the results.
        :param remote: Remote name to be added
        :param uri: Remote address to be added
        :return:
        """
        raise NotImplementedError("")

    @property
    def uri(self):
        return self._uri

    @property
    def pypads(self):
        return self._pypads

    @property
    def managed_result_git(self):
        return self._managed_result_git

    @abstractmethod
    def store_tracked_object(self, to):
        raise NotImplementedError("")

    @abstractmethod
    def store_logger_output(self, lo):
        raise NotImplementedError("")

    @abstractmethod
    def load_tracked_object(self, path):
        raise NotImplementedError("")

    @abstractmethod
    def log_artifact(self, local_path, meta: ArtifactMetaModel, artifact_path=None):
        raise NotImplementedError("")

    @abstractmethod
    def log_mem_artifact(self, artifact, meta: ArtifactMetaModel):
        raise NotImplementedError("")

    @abstractmethod
    def log_metric(self, metric, meta: MetricMetaModel):
        raise NotImplementedError("")

    @abstractmethod
    def log_parameter(self, parameter, meta: ParameterMetaModel):
        raise NotImplementedError("")

    @abstractmethod
    def set_tag(self, tag, meta: TagMetaModel):
        raise NotImplementedError("")


class MLFlowBackend(BackendInterface):
    """
    Backend pushing data to mlflow
    """

    def __init__(self, uri, pypads):
        """
        :param uri: Location in which we want to write results.
        :param pypads: Owning pypads instance
        :return:
        """
        super().__init__(uri, pypads)
        # Set the tracking uri
        mlflow.set_tracking_uri(self._uri)

    def manage_results(self, result_path):
        self._managed_result_git = self.pypads.managed_git_factory(result_path)

        def commit(pads, *args, **kwargs):
            message = "Added results for run " + pads.api.active_run().info.run_id
            pads.managed_result_git.commit_changes(message=message)

            repo = pads.managed_result_git.repo
            remotes = repo.remotes

            if not remotes:
                logger.warning(
                    "Your results don't have any remote repository set. Set a remote repository for"
                    "to enable automatic pushing.")
            else:
                for remote in remotes:
                    name, url = remote.name, list(remote.urls)[0]
                    try:
                        # check if remote repo is bare and if it is initialize it with a temporary local repo
                        pads.managed_result_git.is_remote_empty(remote=name,
                                                                remote_url=url,
                                                                init=True)
                        # stash current state
                        repo.git.stash('push', '--include-untracked')
                        # Force pull
                        repo.git.pull(name, 'master', '--allow-unrelated-histories')
                        # Push merged changes
                        repo.git.push(name, 'master')
                        logger.info("Pushed your results automatically to " + name + " @:" + url)
                        # pop the stash
                        repo.git.stash('pop')
                    except Exception as e:
                        logger.error("pushing logs to remote failed due to this error '{}'".format(str(e)))

        self.pypads.api.register_teardown_fn("commit", commit, nested=False, intermediate=False,
                                             error_message="A problem executing the result management function was detected."
                                                           " Check if you have to commit / push results manually."
                                                           " Following exception caused the problem: {0}",
                                             order=sys.maxsize - 1)

    def add_result_remote(self, remote, uri):
        if self.managed_result_git is None:
            raise Exception("Can only add remotes to the result directory if it is managed by pypads git.")
        try:
            self.managed_result_git.remote = remote
            self.managed_result_git.remote_uri = uri
            self.managed_result_git.repo.create_remote(remote, uri)
        except Exception as e:
            logger.warning("Failed to add remote due to exception: " + str(e))

    @property
    def mlf(self) -> MlflowClient:
        return MlflowClient(self.uri)

    def store_tracked_object(self, to: TrackedObject, path=""):
        path += "{}#{}".format(to.__class__.__name__, id(to))
        try_write_artifact(path, to.json(), write_format=WriteFormats.json)
        return path

    def store_logger_output(self, lo: LoggerOutput, path=""):
        path += "{}/{}".format("Output", id(lo))
        try_write_artifact(path, lo.json(), write_format=WriteFormats.json)
        return path

    def log_artifact(self, local_path, meta: ArtifactMetaModel, artifact_path=None):
        if artifact_path is None:
            if meta:
                artifact_path = meta.path
        try_mlflow_log(mlflow.log_artifact, local_path, artifact_path)

    def log_mem_artifact(self, artifact, meta: Union[MetadataModel, ArtifactMetaModel], preserve_folder=True):
        try_write_artifact(meta.path, artifact, write_format=meta.format, preserve_folder=preserve_folder)

    def log_metric(self, metric, meta: MetricMetaModel):
        mlflow.log_metric(meta.name, metric, meta.step)

    def log_parameter(self, parameter, meta: ParameterMetaModel):
        mlflow.log_param(meta.name, parameter)

    def set_tag(self, tag, meta: TagMetaModel):
        mlflow.set_tag(meta.name, tag)



import inspect
from typing import Set

from pypads import logger
from pypads.importext.mappings import MatchedMapping
from pypads.importext.wrapping.base_wrapper import BaseWrapper, Context


class ModuleWrapper(BaseWrapper):

    def __init__(self, pypads):
        super().__init__(pypads)
        self._punched_module_names = set()

    @property
    def punched_module_names(self):
        return self._punched_module_names

    def add_punched_module_name(self, name):
        self._punched_module_names.add(name)

    def wrap(self, module, context, matched_mappings: Set[MatchedMapping]):
        """
        Function to wrap modules with pypads functionality
        :param module:
        :param context:
        :param matched_mappings:
        :return:
        """
        if module.__name__ not in self.punched_module_names:
            self._pypads.wrap_manager.module_wrapper.add_punched_module_name(module.__name__)
            for matched_mapping in matched_mappings:
                context.store_wrap_meta(matched_mapping, module)

            if not context.has_original(module):
                context.store_original(module)

            # Try to wrap every attr of the module
            # Only get entries defined directly in this module
            # https://stackoverflow.com/questions/22578509/python-get-only-classes-defined-in-imported-module-with-dir
            attrs = {}
            for matched_mapping in matched_mappings:
                for name in list(filter(matched_mapping.mapping.applicable_filter(
                        Context(module, ".".join([context.reference, module.__name__]))),
                        [m for m, _ in inspect.getmembers(module, lambda x: hasattr(x,
                                                                                    "__module__") and x.__module__ == module.__name__)])):
                    attr = getattr(module, name)
                    if attr not in attrs:
                        attrs[attr] = set()
                    else:
                        attrs[attr].add(matched_mapping)

            for attr, mm in attrs.items():
                # Don't track imported modules
                if not inspect.ismodule(attr):
                    self._pypads.wrap_manager.wrap(attr, module, mm)
        else:
            logger.debug("Module " + str(module) + " already duck-punched.")
        return module



import os
import threading
from typing import Type

from pydantic import BaseModel

from pypads import logger
from pypads.importext.wrapping.base_wrapper import Context
from pypads.model.metadata import ModelObject
from pypads.model.models import FunctionReferenceModel, CallAccessorModel, CallIdModel, CallModel


class FunctionReference(ModelObject):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return FunctionReferenceModel

    def __init__(self, _pypads_context: Context, _pypads_wrappee, *args, **kwargs):
        self.wrappee = _pypads_wrappee
        super().__init__(*args, context=_pypads_context, fn_name=_pypads_wrappee.__name__,
                         **kwargs)
        self._real_context = None
        self._function_type = None

        if self.is_wrapped():
            self.wrappee = self.context.container.__dict__[self.wrappee.__name__]

    def real_context(self):
        """
        Find where the accessor function was defined
        :return:
        """

        # Return if already found
        if self._real_context:
            return self._real_context
        self._real_context = self.context.real_context(self.wrappee.__name__)
        return self._real_context

    def function_type(self):
        """
        Get the function type of the accessor function.
        :return:
        """

        # Return if already found
        if self._function_type:
            return self._function_type

        if self.context.is_module():
            function_type = "staticmethod"
        else:
            # Get the function type (Method, unbound etc.)
            try:
                real_ctx = self.real_context()
                if real_ctx is None:
                    raise ValueError("Couldn't find real context.")
                function_type = type(real_ctx.get_dict()[self.wrappee.__name__])
            except Exception as e:
                logger.warning("Couldn't get function type of '" + str(self.wrappee.__name__) + "' on '" + str(
                    self.real_context()) + ". Omit logging. " + str(e))
                return None

            # TODO Can we find less error prone ways to get the type of the given fn?
            # Delegate decorator of sklearn obfuscates the real type.
            # if is_package_available("sklearn"):
            #     from sklearn.utils.metaestimators import _IffHasAttrDescriptor
            #     if function_type == _IffHasAttrDescriptor:
            if str(function_type) == "<class 'sklearn.utils.metaestimators._IffHasAttrDescriptor'>":
                function_type = "wrapped"
                self.wrappee = self.real_context().get_dict()[self.wrappee.__name__]

            # Set cached result
            self._function_type = function_type
        return function_type

    def is_static_method(self):
        return "staticmethod" in str(self.function_type())

    def is_function(self):
        return "function" in str(self.function_type())

    def function_name(self):
        return self.wrappee.__name__

    def is_class_method(self):
        return "classmethod" in str(self.function_type())

    def is_wrapped(self):
        return "wrapped" in str(self.function_type())

    @property
    def function_id(self):
        if hasattr(self.wrappee, "__name__"):
            if hasattr(self.context, self.wrappee.__name__):
                return id(getattr(self.context, self.wrappee.__name__))
        return str(id(self.context)) + "." + str(id(self.wrappee))

    def __str__(self):
        return str(self._real_context) + "." + str(self.wrappee.__name__)


class CallAccessor(FunctionReference):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return CallAccessorModel

    def __init__(self, *args, instance, _pypads_context, _pypads_wrappee, **kwargs):
        super().__init__(_pypads_context, _pypads_wrappee, instance_id=id(instance),
                         **kwargs)
        self._instance = instance

    @property
    def instance(self):
        return self._instance

    @classmethod
    def from_function_reference(cls, function_reference: FunctionReference, instance):
        return CallAccessor(instance=instance, _pypads_context=function_reference.context,
                            _pypads_wrappee=function_reference.wrappee)

    def is_call_identity(self, other):
        if other.is_class_method() or other.is_static_method() or other.is_wrapped():
            if other.context == self.context:
                if other.wrappee.__name__ == self.wrappee.__name__:
                    return True
        if other.is_function():
            if other.instance_id == self.instance_id:
                if other.wrappee.__name__ == self.wrappee.__name__:
                    return True


class CallId(CallAccessor):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return CallIdModel

    def __init__(self, instance, _pypads_context,
                 _pypads_wrappee, instance_number, call_number, **kwargs):
        super().__init__(instance=instance, _pypads_context=_pypads_context, _pypads_wrappee=_pypads_wrappee,
                         process=os.getpid(), thread=threading.get_ident(),
                         instance_number=instance_number, call_number=call_number,
                         **kwargs)

    @classmethod
    def from_accessor(cls, accessor: CallAccessor, instance_number, call_number):
        return CallId(accessor.instance, accessor.context, accessor.wrappee, instance_number, call_number)

    def to_parent_folder(self):
        return os.path.join("process_" + str(self.process) + str(self.thread))

    def to_folder(self):
        return os.path.join(*self.to_fragements())

    def __str__(self):
        return ".".join(self.to_fragements())

    def to_fragements(self):
        return ("process_" + str(self.process), "thread_" + str(self.thread),
                "context_" + self.context.container.__name__,
                "instance_" + str(
                    self.instance_number), "function_" + self.wrappee.__name__, "call_" + str(self.call_number))


class Call(ModelObject):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return CallModel

    def __init__(self, call_id: CallId, *args, **kwargs):
        super().__init__(*args, call_id=call_id, **kwargs)
        self._active_hooks = set()

    def finish(self):
        self.finished = True

    def add_hook(self, hook):
        self._active_hooks.add(hook)

    def has_hook(self, hook):
        return hook in self._active_hooks

    def remove_hook(self, hook):
        self._active_hooks.remove(hook)

    def to_folder(self):
        return self.call_id.to_folder()



anchors = {}


class Anchor:
    """
    Class defining a name, description etc. on which a hook is based.
    """

    def __init__(self, name, description):
        """
        Constructor for an anchor.
        :param name: Name of the anchor.
        :param description: String describing the anchor and its purpose.
        """
        self._name = name
        self._description = description
        if self._name in anchors:
            raise Exception("Anchor with name {} already exists".format(self._name))
        anchors[self._name] = self

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        """
        Return a string describing the anchor and its purpose.
        :return:
        """
        return self._description

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return self._name


def init_anchors():
    if not all([a.name in anchors for a in DEFAULT_ANCHORS]):
        raise Exception("There seems to be an issues with adding the anchors")


def get_anchor(name):
    if name not in anchors:
        return None
    return anchors.get(name)


# TODO maybe change to enum
DEFAULT_ANCHORS = [Anchor("pypads_init", "Used if a tracked concept is initialized."),
                   Anchor("pypads_fit", "Used if an model is fitted to data."),
                   Anchor("pypads_predict", "Used if an model predicts something."),
                   Anchor("pypads_metric", "Used if an metric is compiled."),
                   Anchor("pypads_log", "Used to only log a call.")]

init_anchors()



import os

from pypads import logger
from pypads.app.misc.mixins import DefensiveCallableMixin, DependencyMixin


class ManagedGitFactory(DefensiveCallableMixin, DependencyMixin):
    """
    Class creating a managed git on path. A Managed git is used to interact with a git repository to preserve changes
    by branching etc.
    """

    def __init__(self, pads, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pads = pads

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        logger.warning("Couldn't initialized git repository because of exception: {0}".format(error))

    _dependencies = {"git"}

    def __real_call__(self, path, *args, **kwargs):
        return ManagedGit(path, pads=self.pads)


class ManagedGit:
    """
     Managed git on local disk. A Managed git is used to interact with a git repository to preserve changes by
     branching etc.
    """

    def __init__(self, path, pads=None):
        import git

        if not pads:
            from pypads.app.base import get_current_pads
            self.pads = get_current_pads()
        else:
            self.pads = pads

        # if path is a file we init on the parent directory
        if os.path.isfile(path):
            path = os.path.dirname(path)
        # if path is not representing the source code path
        path = self._verify_path(path)
        from git import InvalidGitRepositoryError
        try:
            self.repo = git.Repo(path, search_parent_directories=True)
        except InvalidGitRepositoryError:
            logger.warning("No existing git repository was found on {0}, initializing a new one...".format(path))
            self._init_git_repo(path)

    def _verify_path(self,path, pads=None):
        # Fix: when using PyPads within a IPython Notebook.
        if path != os.getcwd():
            path = os.getcwd()
            if pads:
                from mlflow.utils.mlflow_tags import MLFLOW_SOURCE_NAME
                pads.api.set_tag(MLFLOW_SOURCE_NAME, path)
        return path

    def _init_git_repo(self, path):
        import git
        from git import InvalidGitRepositoryError, GitCommandError, GitError
        try:
            self.repo = git.Repo.init(path, bare=False)
            self._add_git_ignore()
            logger.info("Repository was successfully initialized")
        except (InvalidGitRepositoryError, GitCommandError, GitError) as e:
            raise Exception(
                "No repository was present and git could not initialize a repository in this directory"
                " {0} because of exception: {1}".format(path, e))

    def preserve_changes(self, message=""):
        try:
            orig_branch = self.repo.active_branch.name
            run = self.pads.api.active_run()
            if len(self.repo.index.diff('HEAD')) > 0 or len(
                    self.repo.index.diff(None)) > 0 or self.repo.untracked_files:
                logger.warning("There are uncommitted changes in your git!")
                # Save those changes to stash
                self.repo.git.stash('push', '--include-untracked')

                # check if the changes were already tracked by PyPads
                branch, diff = self.search_tracking_branch(ref=orig_branch)

                if not branch:
                    branch, diff = self.create_tracking_branch(message)
                    logger.info("Created branch " + branch.name)
                    # Log the commit hash, branch and diff
                else:
                    logger.info("Using already existing pypads branch " + branch.name)

                # checkout to the original branch
                self.repo.git.checkout(orig_branch)

                # and pop the stash
                self.repo.git.stash('pop')
            else:
                branch = orig_branch
                diff = None
            return branch, diff
        except Exception as e:
            raise Exception("Preserving commit failed due to %s" % str(e))

    def search_tracking_branch(self, ref='HEAD'):
        """
        Compares the current untracked changes to all existing pypads managed branches.
        Returns the branch if we tracked the changes already in it, otherwise None.
        :return:
        """
        # get the branches created by PyPads
        branches = {b.name: b for b in self.repo.branches if "PyPads" in b.name}
        # get the tags saved by PyPads
        tags = [t.tag for t in self.repo.tags if 'PyPads' in t.path]
        try:
            for tag in tags:
                diff = self.repo.git.diff(ref, tag.tag, '--raw')
                if diff == tag.message:
                    return branches.get(tag.tag), self.repo.git.diff(ref, tag.tag)
        except Exception as e:
            logger.warning("Checking existing branches failed due to %s" % str(e))
            return None, None
        return None, None

    def create_tracking_branch(self, message):
        orig_branch = self.repo.active_branch.name
        run = self.pads.api.active_run()
        logger.warning("Stashing, branching out, "
                       "committing, reverting back and unstashing...")
        # branch out, apply the stashed changes and commit
        branch_name = "PyPads/{}".format(run.info.run_id)
        branch = self.repo.git.checkout(orig_branch, b=branch_name)

        self.repo.git.stash('apply')
        self.add_untracked_files()
        self.repo.git.commit(message=message)

        # create the tag with diff for this branch
        diff_raw = self.repo.git.diff('master', '--raw')
        diff = self.repo.git.diff('master')
        # TODO hash diff?
        self.repo.create_tag(path=branch_name, message=diff_raw)
        return branch, diff

    def commit_changes(self, message=""):
        try:
            self.add_untracked_files()
            self._commit(message)
        except Exception as e:
            raise Exception("Failed to commit due to following exception: %s" % str(e))

    def _commit(self, message=""):
        self.repo.git.commit(message=message)
        self.pads.api.set_tag("pypads.source.git.commit", self.repo.head.object.hexsha)
        self.pads.api.set_tag("pypads.git.branch", self.repo.active_branch.name)

    def add_untracked_files(self):
        untracked_files = self.repo.untracked_files
        if untracked_files and len(untracked_files) > 0:
            self.repo.index.add(untracked_files)
        if len([item.a_path for item in self.repo.index.diff(None)]) > 0:
            self.repo.git.add(A=True)

    def is_remote_empty(self, remote="", remote_url="", init=False):
        from tempfile import TemporaryDirectory
        from git import Repo
        with TemporaryDirectory() as temp_dir:
            repo = Repo.clone_from(remote_url, temp_dir)
            if not repo.branches and init:
                with open(temp_dir + "/Readme.md", "w") as f:
                    f.write("# Results repository")
                repo.git.add(A=True)
                repo.git.commit(message="Initializing the repository with Readme.md")
                repo.git.push(remote, 'master')

    def _add_git_ignore(self):
        try:
            with open(self.repo.working_dir + "/.gitignore", "w") as file:
                file.write(GIT_IGNORE)
            self.repo.git.add(A=True)
        except Exception as e:
            logger.warning("Could add .gitignore file to the repo due to this %s" % str(e))


GIT_IGNORE = """
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
.pybuilder/
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
#   For a library or package, you might want to ignore these files since the code is
#   intended to run in multiple environments; otherwise, check them in:
# .python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# PEP 582; used by e.g. github.com/David-OConnor/pyflow
__pypackages__/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/

# Cython debug symbols
cython_debug/
"""



event_types = {}


class EventType:
    """
    Pypads event type. Every event type triggers multiple logging functions to be executed.
    Events are triggered by hooks.
    """

    def __init__(self, name, description):
        """
        Constructor for an event.
        :param name: Name of the event.
        :param description: Description of the event.
        """
        self._name = name
        self._description = description
        event_types[self._name] = self

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description


def init_event_types():
    if not all([a.name in event_types for a in DEFAULT_EVENT_TYPES]):
        raise Exception("There seems to be an issues with adding the anchors")


def get_event_type(name):
    if name not in event_types:
        return None
    return event_types.get(name)


# TODO maybe change to enum
DEFAULT_EVENT_TYPES = [EventType("parameters", "Track the parameters for given model."),
                       EventType("output", "Track the output of the function."),
                       EventType("input", "Track the input of the function."),
                       EventType("hardware", "Track current hardware load on function execution."),
                       EventType("metric", "Track a metric."),
                       EventType("autolog", "Activate mlflow autologging."),
                       EventType("pipeline", "Track a pipeline step."),
                       EventType("log", "Log the call to console."),
                       EventType("init", "Log the tracked class init to console.")]

init_event_types()



from abc import ABCMeta
from functools import wraps
from typing import Type

from pydantic import BaseModel, HttpUrl

from pypads.app.injections.base_logger import LoggerCall, SimpleLogger
from pypads.app.misc.extensions import ExtendableMixin, Plugin
from pypads.injections.analysis.determinism import check_determinism
from pypads.model.models import LoggerModel
from pypads.utils.util import inheritors

validator_plugins = set()


class Validator(SimpleLogger, metaclass=ABCMeta):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/validator"

    def build_call_object(self, _pypads_env, **kwargs):
        return LoggerCall(logging_env=_pypads_env,
                          is_a="https://www.padre-lab.eu/onto/ValidatorLoggerCall", **kwargs)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return LoggerModel

    def _base_path(self):
        return "Validators/{}/".format(self.__class__.__name__)


class IValidators(Plugin):

    def __init__(self):
        super().__init__(type=Validator)
        validator_plugins.add(self)

    def _get_meta(self):
        """ Method returning information about where the actuator was defined."""
        return self.__module__

    def _get_methods(self):
        return [method_name for method_name in dir(self) if callable(getattr(object, method_name))]


def validator(f):
    """
    Validator used to convert a function to a tracked actuator.
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        # self is an instance of the class
        return Validator(fn=f)(self, *args, **kwargs)

    return wrapper


class PyPadsValidators(IValidators):
    def __init__(self):
        super().__init__()

    @property
    def pypads(self):
        from pypads.app.pypads import get_current_pads
        return get_current_pads()

    @validator
    def determinism(self):
        check_determinism()


class ValidatorPluginManager(ExtendableMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(plugin_list=validator_plugins)


pypads_validators = PyPadsValidators()


def validators():
    """
    Returns classes of
    :return:
    """
    return inheritors(Validator)



from pydantic.main import BaseModel
from pydantic.networks import HttpUrl
from typing import List, Type

from app.env import LoggerEnv
from pypads.app.injections.base_logger import TrackedObject, LoggerCall
from pypads.app.injections.run_loggers import RunSetup
from pypads.model.models import TagMetaModel, TrackedObjectModel, OutputModel
from pypads.utils.util import sizeof_fmt, local_uri_to_path


class HardwareTO(TrackedObject):
    """
    Tracking object class for System info, i.e: cpu, os, memory, disk
    """

    class HardwareModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/env/hardware-information"
        name: str = "Hardware Info"
        report: List[TagMetaModel] = []

        class Config:
            orm_mode = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.HardwareModel

    def __init__(self, *args, tracked_by: LoggerCall, name: str, uri: str, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, name=name, uri=uri, **kwargs)

    def add_tag(self, name, value, description):
        meta = TagMetaModel(name=name, description=description)
        self.report.append(meta)

        self._store_tag(value, meta)


class ISystemRSF(RunSetup):
    _dependencies = {"psutil"}

    name = "System Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/system-run-logger"

    class ISystemRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/ISystemRSF-Output"
        system_info: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.ISystemRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        import platform
        uname = platform.uname()
        system_info = HardwareTO(name="System Info", tracked_by=_logger_call,
                                 uri="https://www.padre-lab.eu/onto/env/system-information")

        system_info.add_tag("pypads.system", uname.system, description="Operating system")
        system_info.add_tag("pypads.system.node", uname.node, description="Operating system node")
        system_info.add_tag("pypads.system.release", uname.release, description="Operating system release")
        system_info.add_tag("pypads.system.version", uname.version, description="Operating system version")
        system_info.add_tag("pypads.system.machine", uname.machine, description="Operating system machine")
        system_info.add_tag("pypads.system.processor", uname.processor, description="Processor technology")
        system_info.store(_logger_output, "system_info")


class ICpuRSF(RunSetup):
    _dependencies = {"psutil"}
    name = "CPU Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/cpu-run-logger"

    class ICpuRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/ICpuRSF-Output"
        cpu_info: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.ICpuRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        import psutil
        cpu_info = HardwareTO(name="Cpu Info", tracked_by=_logger_call,
                              uri="https://www.padre-lab.eu/onto/env/cpu-information")
        cpu_info.add_tag("pypads.system.cpu.physical_cores", psutil.cpu_count(logical=False),
                         description="Number of physical cores")
        cpu_info.add_tag("pypads.system.cpu.total_cores", psutil.cpu_count(logical=True),
                         description="Number of total cores")
        freq = psutil.cpu_freq()
        cpu_info.add_tag("pypads.system.cpu.max_freq", f"{freq.max:2f}Mhz",
                         description="Maximum processor frequency in (Mhz)")
        cpu_info.add_tag("pypads.system.cpu.min_freq", f"{freq.min:2f}Mhz",
                         description="Minimum processor frequencyin (Mhz)")
        cpu_info.store(_logger_output, "cpu_info")


class IRamRSF(RunSetup):
    _dependencies = {"psutil"}
    name = "Ram Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/Ram-run-logger"

    class IRamRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/IRamRSF-Output"
        memory_info: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.IRamRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        memory_info = HardwareTO(name="Memory Info", tracked_by=_logger_call,
                                 uri="https://www.padre-lab.eu/onto/env/memory-information")
        import psutil
        memory = psutil.virtual_memory()
        memory_info.add_tag("pypads.system.memory.total", sizeof_fmt(memory.total),
                            description="Total virtual memory RAM")
        swap = psutil.swap_memory()
        memory_info.add_tag("pypads.system.swap.total", sizeof_fmt(swap.total), description="Total swap memory")
        memory_info.store(_logger_output, "memory_info")


class IDiskRSF(RunSetup):
    _dependencies = {"psutil"}
    name = "Disk Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/disk-run-logger"

    class IDiskRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/IDiskRSF-Output"
        disk_info: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.IDiskRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        disk_info = HardwareTO(name="Disk Info", tracked_by=_logger_call,
                               uri="https://www.padre-lab.eu/onto/env/disk-information")
        import psutil
        # see https://www.thepythoncode.com/article/get-hardware-system-information-python
        pads = _logger_call._logging_env.pypads
        path = local_uri_to_path(pads.backend.uri)
        disk_usage = psutil.disk_usage(path)
        disk_info.add_tag("pypads.system.disk.total", sizeof_fmt(disk_usage.total), description="Total disk usage")
        disk_info.store(_logger_output, "disk_info")


class IPidRSF(RunSetup):
    _dependencies = {"psutil"}
    name = "Process Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/process-run-logger"

    class IPidRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/IPidRSF-Output"
        process_info: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.IPidRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        process_info = HardwareTO(name="Process Info", tracked_by=_logger_call,
                                  uri="https://www.padre-lab.eu/onto/env/process-information")
        import psutil
        import os
        pid = os.getpid()
        process = psutil.Process(pid=pid)
        process_info.add_tag("pypads.system.process.id", pid, description="Process ID (PID)")
        process_info.add_tag("pypads.system.process.cwd", process.cwd(),
                             description="Process current working directory")
        process_info.add_tag("pypads.system.process.cpu_usage", str(process.cpu_percent()) + "%",
                             description="Process cpu usage")
        process_info.add_tag("pypads.system.process.memory_usage", str(process.memory_percent()) + "%",
                             description="Process memroy usage")
        process_info.store(_logger_output, "process_info")


class ISocketInfoRSF(RunSetup):
    name = "Socket Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/socket-run-logger"

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.ISocketInfoRSFOutput

    class ISocketInfoRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/ISocketInfoRSF-Output"
        socket_info: HardwareTO.get_model_cls() = ...

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        socket_info = HardwareTO(name="Socket Info", tracked_by=_logger_call,
                                 uri="https://www.padre-lab.eu/onto/env/socker-information")
        import socket
        socket_info.add_tag("pypads.system.hostname", socket.gethostname(), description="Hostname of open socket")
        socket_info.add_tag("pypads.system.ip-address", socket.gethostbyname(socket.gethostname()),
                            description="Ip address of open socket")
        socket_info.store(_logger_output, "socket_info")


class IMacAddressRSF(RunSetup):
    name = "MacAddress Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/macaddress-run-logger"

    class IMacAddressRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/IMacAddressRSF-Output"
        mac_address: HardwareTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.IMacAddressRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        mac_address = HardwareTO(name="Mac Address", tracked_by=_logger_call,
                                 uri="https://www.padre-lab.eu/onto/env/mac-address-information")
        import re, uuid
        mac_address.add_tag("pypads.system.macaddress", ':'.join(re.findall('..', '%012x' % uuid.getnode())),
                            description="Mac Address")
        mac_address.store(_logger_output, "mac_address")

# def inetw(pads):
#     if is_package_available("psutil"):
#         import psutil
#         # get net stats
#
#     else:
#         logger.warning("To track network usage you need to install psutil.")



import random
from logging import warning
from padrepads.util import _is_package_available

# Logging

def log_random_seed(key):
    from pypads.app.base import PyPads
    from pypads.app.pypads import get_current_pads
    pads: PyPads = get_current_pads()

    # Get seed information from cache
    if pads.cache.run_exists(key):
        # TODO if tag already exists (set called multiple times) we need to handle that (save seed per function run?)
        pads.api.set_tag("pypads." + key, pads.cache.run_get(key))
    else:
        warning("Can't log seed produced by seed generator. You have to enable ")


# --- random seed ---
original_random = random.seed


def random_seed(seed):
    try:
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        pads.cache.run_add("random.seed", seed)
        log_random_seed("random.seed")
        return original_random(seed)
    except Exception as e:
        Warning("Tracker failed to log the set seed because %s" % str(e))
        return original_random(seed)


random.seed = random_seed

# --- numpy seed ---
# noinspection PyPackageRequirements,PyUnresolvedReferences
import numpy

original_numpy = numpy.random.seed


def numpy_seed(seed):
    try:
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        pads.cache.run_add("numpy.random.seed", seed)
        log_random_seed("numpy.random.seed")
        return original_numpy(seed)
    except Exception as e:
        Warning("Tracker failed to log the set seed because %s" % str(e))
        return original_numpy(seed)


numpy.random.seed = numpy_seed

# --- pytorch seed ---
if _is_package_available("torch"):
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    import torch

    original_torch = torch.manual_seed


    def torch_seed(seed):
        try:
            from pypads.app.pypads import get_current_pads
            pads = get_current_pads()
            pads.cache.run_add("torch.seed", seed)
            log_random_seed("torch.seed")
            return original_torch(seed)
        except Exception as e:
            Warning("Tracker failed to log the set seed because %s" % str(e))
            return original_torch(seed)


    torch.manual_seed = torch_seed

    if torch.cuda.is_available():
        original_torch_cuda = torch.cuda.manual_seed_all


        def torch_cuda_seed(seed):
            try:
                from pypads.app.pypads import get_current_pads
                pads = get_current_pads()
                pads.cache.run_add("torch.cuda.seed", seed)
                return original_torch_cuda(seed)
            except Exception as e:
                Warning("Tracker failed to log the set seed because %s" % str(e))
                return original_torch_cuda(seed)


        torch.cuda.manual_seed_all = torch_cuda_seed


def set_random_seed(seed):
    import random
    global padre_seed
    padre_seed = seed

    # --- set random seed ---
    random.seed(seed)

    # --- set numpy seed ---
    numpy.random.seed(seed)
    # global seeds for numpy seem to not work with RandomState()

    # --- set pytorch seed ---
    if _is_package_available("torch"):
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import torch
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)



from typing import Iterable, Union

from pypads.bindings.event_types import EventType
from pypads.bindings.hooks import Hook
from pypads.importext.versioning import LibSelector
from pypads.injections.analysis.parameters import ParametersILF
from pypads.injections.loggers.data_flow import OutputILF, InputILF
from pypads.injections.loggers.debug import Log, LogInit
from pypads.injections.loggers.hardware import CpuILF, RamILF, DiskILF
from pypads.injections.loggers.metric import MetricILF
from pypads.injections.loggers.mlflow.mlflow_autolog import MlflowAutologger
from pypads.injections.loggers.pipeline_detection import PipelineTrackerILF
from pypads.utils.logging_util import WriteFormats

# maps events to loggers
# Default event mappings. We allow to log parameters, output defor input
DEFAULT_LOGGING_FNS = {
    "parameters": ParametersILF(),
    "output": OutputILF(_pypads_write_format=WriteFormats.text),
    "input": InputILF(_pypads_write_format=WriteFormats.text),
    "hardware": [CpuILF(_pypads_write_format=WriteFormats.text), RamILF(_pypads_write_format=WriteFormats.text),
                 DiskILF(_pypads_write_format=WriteFormats.text)],
    "metric": MetricILF(),
    "autolog": MlflowAutologger(),
    "pipeline": PipelineTrackerILF(_pypads_pipeline_type="normal", _pypads_pipeline_args=False),
    "log": Log(),
    "init": LogInit()
}


class Event:
    """
    Pypads event. Every event triggers multiple logging functions to be executed. Events are triggered by hooks.
    """

    def __init__(self, event_type: EventType, triggered_by: Hook):
        """
        Constructor for an event.
        :param event_type: Type of the event.
        :param triggered_by: Source which triggered the event.
        """
        self._event_type = event_type
        self._source = triggered_by

    @property
    def name(self):
        return self._event_type.name

    @property
    def description(self):
        return self._event_type.description

    @property
    def source(self):
        return self._source


class FunctionRegistry:

    def __init__(self, pypads):
        self._pypads = pypads
        self._fns = {}

    @staticmethod
    def from_dict(pypads, mapping):
        """
        Build a function registry from given dict mapping.
        :param mapping: A dict containing the mapping information.
        :return:
        """
        registry = FunctionRegistry(pypads)

        if mapping is None:
            mapping = DEFAULT_LOGGING_FNS

        for key, value in mapping.items():
            if isinstance(value, Iterable):
                registry.add_function(key, *value)
            elif callable(value):
                registry.add_function(key, value)
        return registry

    def add_function(self, event_name: str, *fns):
        """
        Add all given functions behind a key
        :param event_name: Event key for the functions
        :param fns: Functions to add
        :return:
        """
        if event_name not in self._fns:
            self._fns[event_name] = set()
        for fn in fns:
            self._fns[event_name].add(fn)

    def has(self, event_name: Union[str,tuple]):
        """
        Check if at least one function with event key is in map.
        :param event_name: Key of the function
        :return: true if has function
        """
        return event_name in self._fns

    def get_functions(self, event_name, lib_selector: LibSelector):
        if not self.has(event_name):
            return set()
        fns = self._fns[event_name] if isinstance(self._fns[event_name], Iterable) else [self._fns[event_name]]

        fitting_fns = []
        for fn in fns:
            fitting_fns = fitting_fns + [(lib.specificity, fn) for lib in fn.supported_libraries if
                                         lib.allows_any(lib_selector)]

        # Filter for specificity and identity
        identities = {}
        filtered_fns = set()
        for spec, fn in fitting_fns:
            if fn.uid is None:
                filtered_fns.add(fn)
            elif fn.uid in identities:
                # If we are more specific and have the same identity remove old fn
                if identities[fn.uid][0] <= spec:
                    if identities[fn.uid][0] < spec:
                        if fn.uid in identities:
                            filtered_fns.remove(identities[fn.uid][1])
                    filtered_fns.add(fn)
                    identities[fn.uid] = (spec, fn)
            else:
                # If not seen add it
                filtered_fns.add(fn)
                identities[fn.uid] = (spec, fn)
        return filtered_fns



from typing import Iterable, Set

from pypads.app.misc.mixins import OrderMixin, DEFAULT_ORDER
from pypads.utils.logging_util import WriteFormats

# Maps hooks to events

DEFAULT_HOOK_MAPPING = {
    "init": {"on": ["pypads_init"]},
    "parameters": {"on": ["pypads_fit"]},
    "hardware": {"on": ["pypads_fit"]},
    "output": {"on": ["pypads_fit", "pypads_predict"]},
    "input": {"on": ["pypads_fit"], "with": {"_pypads_write_format": WriteFormats.text.name}},
    "metric": {"on": ["pypads_metric"]},
    "pipeline": {"on": ["pypads_fit", "pypads_predict", "pypads_transform", "pypads_metric"]},
    "log": {"on": ["pypads_log"]}
}


class Hook:
    """
    Pypads hooks. Hooks are defined on function executions and trigger events to call logger functions.
    """

    def __init__(self, anchor, mapping):  # type: (Anchor, Mapping) -> Hook
        """
        Constructor for an hook.
        :param anchor: Anchor of the hook.
        :param mapping: Mapping creating the hook call.
        """
        self._anchor = anchor
        self._source = mapping

    @property
    def anchor(self):
        return self._anchor

    @property
    def source(self):  # type: () -> Mapping
        return self._source

    @property
    def library(self):  # type: () -> LibSelector
        return self._source.library

    def __eq__(self, other):
        return self.anchor == other.anchor

    def __hash__(self):
        return hash(self.anchor)

    def __str__(self):
        return "Hook[anchor=" + str(self.anchor) + ", lib=" + str(self.library) + "]"


class HookEventConfig(OrderMixin):
    def __init__(self, hook, event_name, parameters=None, *args, **kwargs):
        self._hook = hook
        self._event_name = event_name
        self._parameters = parameters
        super().__init__(*args, **kwargs)

    @property
    def anchor(self):
        return self._hook

    @property
    def event_name(self):
        return self._event_name

    @property
    def parameters(self):
        return self._parameters


class HookRegistry:
    """
    Class mapping hooks to events.
    """

    def __init__(self, pypads):
        self._pypads = pypads
        self._hook_event_mapping = {}

    def add_reference(self, event_name: str, *hook_names: str, order=DEFAULT_ORDER, parameters=None):
        for hook_name in hook_names:
            if hook_name not in self._hook_event_mapping:
                self._hook_event_mapping[hook_name] = set()
            self._hook_event_mapping[hook_name].add(HookEventConfig(hook_name, event_name, parameters, order=order))

    def get_configs_for_hook(self, hook: Hook) -> Set[HookEventConfig]:
        if hook.anchor.name not in self._hook_event_mapping:
            return set()
        else:
            return self._hook_event_mapping[hook.anchor.name]

    def get_logging_functions(self, *hooks: Hook):
        configs = []
        for hook in hooks:
            configs = configs + [(hook, c) for c in list(self.get_configs_for_hook(hook))]
        configs.sort(key=lambda e: -e[1].order)

        fns = []
        for hook, c in configs:
            found_fns = [(f, c.parameters) for f in
                         self._pypads.function_registry.get_functions(c.event_name, hook.library)]
            found_fns.sort(key=lambda e: -e[0].order)
            fns = fns + found_fns
        return fns

    @staticmethod
    def from_dict(pypads, hook_mapping):
        """
        Build a hook registry from given dict mapping.
        :param pypads: Owning PyPads instance
        :param hook_mapping: A dict containing the mapping information.
        :return:
        """
        registry = HookRegistry(pypads)

        if hook_mapping is None:
            hook_mapping = DEFAULT_HOOK_MAPPING

        for key, value in hook_mapping.items():
            parameters = value["with"] if "with" in value else {}
            hook_names = value["on"]
            order = value["order"] if "order" in value else DEFAULT_ORDER
            if isinstance(hook_names, Iterable):
                registry.add_reference(key, *hook_names, order=order, parameters=parameters)
            else:
                registry.add_reference(key, hook_names, order=order, parameters=parameters)
        return registry



import inspect
import operator
import threading
from functools import reduce

import mlflow
import pkg_resources

from pypads import logger
from pypads.app.misc.caches import Cache


def get_class_that_defined_method(meth):
    """
    Try to find the class / module which defined given method.
    :param meth: Method for which we search an origin.
    :return:
    """
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def dict_merge(*dicts):
    """
    Simple merge of dicts
    :param dicts:
    :return:
    """
    merged = {}
    for d in dicts:
        if isinstance(d, dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    node = merged.setdefault(key, {})
                    merged[key] = dict_merge(node, value)
                else:
                    merged[key] = value
    return merged


def sizeof_fmt(num, suffix='B'):
    """
    Get the mem / disk size in a human readable way.
    :param num:
    :param suffix:
    :return:
    """
    if num == 0:
        return '0'
    import math
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)
    if magnitude > 7:
        return '{:.1f}{}{}'.format(val, 'Yi', suffix)
    return '{:3.1f}{}{}'.format(val, ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z'][magnitude], suffix)


def local_uri_to_path(uri):
    """
    Convert URI to local filesystem path.
    """
    from six.moves import urllib
    path = urllib.parse.urlparse(uri).path if uri.startswith("file:") else uri
    return urllib.request.url2pathname(path)


def string_to_int(s):
    """
    Build a int from a given string.
    :param s:
    :return:
    """
    ord3 = lambda x: '%.3d' % ord(x)
    return int(''.join(map(ord3, s)))


def inheritors(clazz):
    """
    Function getting all subclasses of given class.
    :param clazz: Clazz to search for
    :return:
    """
    subclasses = set()
    unseen = [clazz]
    while unseen:
        parent = unseen.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                unseen.append(child)
    return subclasses


def find_package_regex_versions(regex):
    import pkgutil
    import re
    versions = {
        name: find_package_version(name)
        for finder, name, ispkg
        in pkgutil.iter_modules()
        if re.compile(regex).match(name)
    }
    return versions


def is_package_available(name):
    """
    Check if given package is available.
    :param name: Name of the package
    :return:
    """
    import importlib
    try:
        spam_loader = importlib.util.find_spec(name)
    except Exception as e:
        spam_loader = importlib.find_loader(name)
    return spam_loader is not None


def find_package_version(name: str):
    try:
        import sys
        base_package = sys.modules[name]
        if hasattr(base_package, "__version__"):
            lib_version = getattr(base_package, "__version__")
        else:
            lib_version = pkg_resources.get_distribution(name).version
        return lib_version
    except Exception as e:
        logger.debug("Couldn't get version of package {}".format(name))
        return None


def dict_merge_caches(*dicts):
    """
    Merge two dicts. Entries are overwritten if not mergeable. Cache is supported.
    :param dicts: dicts to merge
    :return:
    """
    merged = {}
    for d in dicts:
        if isinstance(d, dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    node = merged.setdefault(key, {})
                    merged[key] = dict_merge(node, value)
                elif isinstance(value, list):
                    node = merged.setdefault(key, [])
                    try:
                        node.extend(value)
                    except AttributeError as e:
                        try:
                            node = [node]
                            node.extend(value)
                        except Exception as e:
                            logger.error("Failed merging dictionaries in dict_merge_caches : {}".format(str(e)))
                    merged[key] = node
                elif isinstance(value, set):
                    s: set = merged.setdefault(key, set())
                    for v in value:
                        if v in s:
                            merged = dict_merge(v, s.pop(v))
                            s.add(merged)
                        else:
                            s.add(v)
                elif isinstance(value, Cache):
                    node = merged.setdefault(key, Cache())
                    merged[key] = value.merge(node)
                else:
                    merged[key] = value
    return merged


def get_from_dict(d: dict, key_list):
    return reduce(operator.getitem, key_list, d)


def set_in_dict(d: dict, key_list, value):
    get_from_dict(d, key_list[:-1])[key_list[-1]] = value


def has_direct_attr(obj, name):
    """
    Check if self has an attribute
    :param obj: object to check
    :param name: name of the attribute
    :return:
    """
    try:
        object.__getattribute__(obj, name)
        return True
    except AttributeError:
        return False


def get_experiment_id():
    if mlflow.active_run():
        return mlflow.active_run().info.experiment_id
    return None


def get_run_id():
    if mlflow.active_run():
        return mlflow.active_run().info.run_id
    return None


class PeriodicThread(threading.Thread):
    def __init__(self, *args, sleep=1.0, target=None, **kwargs):
        self._stop_event = threading.Event()
        self._sleep_period = sleep
        super().__init__(*args, target=target, **kwargs)

    def run(self):
        try:
            while not self._stop_event.isSet():
                if self._target:
                    self._target(*self._args, **self._kwargs)
                self._stop_event.wait(self._sleep_period)
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs

    def join(self, timeout=None):
        """ Stop the thread. """
        self._stop_event.set()
        threading.Thread.join(self, timeout)



class SuperStop:
    """
    This class resolves the issue TypeError: object.__init__() takes exactly one argument by being the last class
    in a mro and ommitting all arguments. This should be always last in the mro()!
    """

    def __init__(self, *args, **kwargs):
        mro = self.__class__.mro()
        if SuperStop in mro:
            if len(mro) - 2 != mro.index(SuperStop):
                raise ValueError("SuperStop ommitting arguments in " + str(self.__class__)
                                 + " super() callstack: " + str(mro))
        super().__init__()



from typing import Set

from pypads import logger
from pypads.importext.mappings import MatchedMapping
from pypads.importext.wrapping.base_wrapper import BaseWrapper, Context


class ClassWrapper(BaseWrapper):

    def __init__(self, pypads):
        super().__init__(pypads)
        self._punched_class_names = set()

    @property
    def punched_class_names(self):
        return self._punched_class_names

    def wrap(self, clazz, context, matched_mappings: Set[MatchedMapping]):
        """
            Wrap a class in given ctx with pypads functionality
            :param clazz:
            :param context:
            :param matched_mappings:
            :return:
            """
        if clazz.__name__ not in self.punched_class_names:
            for matched_mapping in matched_mappings:
                try:
                    context.store_wrap_meta(matched_mapping, clazz)
                except Exception:
                    return clazz
            self.punched_class_names.add(clazz.__name__)

            if not context.has_original(clazz):
                context.store_original(clazz)

            # Module was changed and should be added to the list of modules which have been changed
            if hasattr(clazz, "__module__"):
                self._pypads.wrap_manager.module_wrapper.add_punched_module_name(clazz.__module__)

            attrs = {}
            clazz_context = Context(clazz, ".".join([context.reference, clazz.__name__]))
            for matched_mapping in matched_mappings:
                # Try to wrap every attr of the class
                for name in list(filter(
                        matched_mapping.mapping.applicable_filter(
                            clazz_context),
                        dir(clazz))):
                    if name not in attrs:
                        attrs[name] = set()
                    attrs[name].add(matched_mapping)

            for name, mm in attrs.items():
                self._pypads.wrap_manager.wrap(getattr(clazz, name), clazz_context, mm)

            # Override class on module
            context.overwrite(clazz.__name__, clazz)
        else:
            logger.debug("Class " + str(clazz) + "already duck-puched.")
        return clazz



import ast
import atexit
import importlib
import os
import pkgutil
from os.path import expanduser
from typing import List

import mlflow

from pypads import logger
from pypads.app.actuators import ActuatorPluginManager
from pypads.app.api import ApiPluginManager
from pypads.app.backends.backend import MLFlowBackend
from pypads.app.decorators import DecoratorPluginManager
from pypads.app.misc.caches import PypadsCache
from pypads.app.validators import ValidatorPluginManager, validators
from pypads.bindings.events import FunctionRegistry
from pypads.bindings.hooks import HookRegistry
from pypads.importext.mappings import MappingRegistry, MappingCollection
from pypads.importext.pypads_import import extend_import_module, duck_punch_loader
from pypads.importext.wrapping.wrapping import WrapManager
from pypads.injections.analysis.call_tracker import CallTracker
from pypads.injections.setup.git import IGitRSF
from pypads.injections.setup.hardware import ISystemRSF, IRamRSF, ICpuRSF, IDiskRSF, IPidRSF, ISocketInfoRSF, \
    IMacAddressRSF
from pypads.injections.setup.misc_setup import DependencyRSF, LoguruRSF

tracking_active = None

# --- Pypads App ---

"""
TODO keras:
Logs loss and any other metrics specified in the fit
    function, and optimizer data as parameters. Model checkpoints
    are logged as artifacts to a 'models' directory.
    EarlyStopping Integration with Keras Automatic Logging
    MLflow will detect if an ``EarlyStopping`` callback is used in a ``fit()``/``fit_generator()``
    call, and if the ``restore_best_weights`` parameter is set to be ``True``, then MLflow will
    log the metrics associated with the restored model as a final, extra step. The epoch of the
    restored model will also be logged as the metric ``restored_epoch``.
    This allows for easy comparison between the actual metrics of the restored model and
    the metrics of other models.
    If ``restore_best_weights`` is set to be ``False``,
    then MLflow will not log an additional step.
    Regardless of ``restore_best_weights``, MLflow will also log ``stopped_epoch``,
    which indicates the epoch at which training stopped due to early stopping.
    If training does not end due to early stopping, then ``stopped_epoch`` will be logged as ``0``.
    MLflow will also log the parameters of the EarlyStopping callback,
    excluding ``mode`` and ``verbose``.
"""

# Default config.
# Pypads mapping files shouldn't interact directly with the logging functions,
# but define events on which different logging functions can listen.
# This config defines such a listening structure.
# {"recursive": track functions recursively. Otherwise check the callstack to only track the top level function.}
DEFAULT_CONFIG = {
    "track_sub_processes": False,
    # Activate to track spawned subprocesses by extending the joblib. This is currently experimental.
    "recursion_identity": False,
    # Activate to ignore tracking on recursive calls of the same function with the same mapping
    "recursion_depth": -1,  # Limit the tracking of recursive calls
    "log_on_failure": True,  # Log the stdout / stderr output when the execution of the experiment failed
    "include_default_mappings": True  # Include the default mappings additionally to the passed mapping if a mapping
    # is passed
}

DEFAULT_SETUP_FNS = {DependencyRSF(), LoguruRSF(), IGitRSF(_pypads_timeout=3), ISystemRSF(), IRamRSF(), ICpuRSF(),
                     IDiskRSF(), IPidRSF(), ISocketInfoRSF(), IMacAddressRSF()}

# Tag name to save the config to in mlflow context.
CONFIG_NAME = "pypads.config"

DEFAULT_EXPERIMENT_NAME = "Default-PyPads"


#  pypads isn't allowed to hold a state anymore (Everything with state should be part of the caching system)
#  - We want to be able to rebuild PyPads from the cache alone if existing
#  to stop the need for pickeling pypads as a whole.

class PyPads:
    """
    PyPads app. Serves as the main entrypoint to PyPads. After constructing this app tracking is activated.
    """

    def __init__(self, uri=None, folder=None, mappings: List[MappingCollection] = None, hooks=None,
                 events=None, setup_fns=None, config=None, pre_initialized_cache: PypadsCache = None,
                 disable_plugins=None, autostart=None, consolidate_outputs=True):
        # Set the singleton instance

        if disable_plugins is None:
            disable_plugins = []
        for name, plugin in discovered_plugins.items():
            if name not in disable_plugins:
                plugin.activate()

        from pypads.app.pypads import set_current_pads
        set_current_pads(self)

        # Init variable to filled later in this constructor
        self._atexit_fns = []

        # Init WrapManager
        self._wrap_manager = WrapManager(self)

        # Init API
        self._api = ApiPluginManager()

        # Init Decorators
        self._decorators = DecoratorPluginManager()

        # Init Actuators
        self._actuators = ActuatorPluginManager()

        # Init Validators
        self._validators = ValidatorPluginManager()

        # Init CallTracker
        self._call_tracker = CallTracker(self)

        # Init cache
        self._cache = pre_initialized_cache if pre_initialized_cache else PypadsCache()

        # Store folder into cache
        self._cache.add("folder", folder or os.path.join(expanduser("~"), ".pypads"))

        # Store uri into cache
        self._cache.add("uri", uri or os.environ.get('MLFLOW_PATH') or os.path.join(self.folder, ".mlruns"))

        # Enable git tracking of the current repository
        from pypads.app.misc.managed_git import ManagedGitFactory
        self._managed_git_factory = ManagedGitFactory(self)

        self._backend = MLFlowBackend(self.uri, self)

        # Store config into cache
        self.config = {**DEFAULT_CONFIG, **config} if config else DEFAULT_CONFIG

        # Store config into cache
        self._cache.add("mappings", mappings)

        # Store hook registry into cache
        self._cache.add("hooks", hooks)

        # Store function registry into cache
        self._cache.add("events", events)

        # Init mapping registry
        self._mapping_registry = MappingRegistry.from_params(self, mappings)

        # Init hook registry
        self._hook_registry = HookRegistry.from_dict(self, hooks)

        # Init function registry
        self._function_registry = FunctionRegistry.from_dict(self, events)

        # Store config into cache
        self._cache.add("mappings", mappings)

        # Store hook registry into cache
        self._cache.add("hooks", hooks)

        # Store function registry into cache
        self._cache.add("events", events)

        # Store a dictionary to consolidate all the output files
        if consolidate_outputs is True:
            self._cache.add("consolidated_dict", dict())

        # Initialize pre run functions before starting a run
        setup_fns = setup_fns or DEFAULT_SETUP_FNS
        for fn in setup_fns:
            self.api.register_setup(fn.__class__.__name__ + "_" + str(id(fn)), fn)

        # Activate tracking by punching the import lib
        if autostart:
            self.activate_tracking()

        # Add cleanup functions
        def cleanup():
            from pypads.app.pypads import get_current_pads
            pads: PyPads = get_current_pads()
            if pads.api.active_run():
                pads.api.end_run()

        self.add_atexit_fn(cleanup)

        if autostart:
            if isinstance(autostart, str):
                self.start_track(autostart)
            else:
                self.start_track()

    @staticmethod
    def available_loggers():
        """
        Return a list of available LoggingFunctions for mappings in the current installation of PyPads.
        :return: Logging function classes
        """
        from pypads.app.injections.injection import logging_functions
        return logging_functions()

    @staticmethod
    def available_setup_functions():
        """
        Return a list of available setup function in the current installation of PyPads.
        :return: Setup function classes
        """
        from pypads.app.injections.run_loggers import run_setup_functions
        return run_setup_functions()

    @staticmethod
    def available_teardown_functions():
        """
        Return a list of available setup function in the current installation of PyPads.
        :return: Teardown function classes
        """
        from pypads.app.injections.run_loggers import run_teardown_functions
        return run_teardown_functions()

    @staticmethod
    def available_anchors():
        """
        Return a list of available anchors in the current installation of PyPads. Anchors are names for hooks.
        :return: Anchors
        """
        from pypads.bindings.anchors import anchors
        return anchors

    @staticmethod
    def available_events():
        """
        Return a list of available events (event types) in the current installation of PyPads. Events are triggered by hooks and map to functions.
        :return: Event types
        """
        from pypads.bindings.event_types import event_types
        return event_types

    @staticmethod
    def available_validators():
        """
        Return a list of available validators in the current installation of PyPads.
        :return: Validator classes
        """
        return validators()

    @staticmethod
    def available_decorators():
        """
        Return a list of available decorators in the current installation of PyPads.
        :return: Decorator classes
        """
        from pypads.app.decorators import decorators
        return decorators()

    @staticmethod
    def available_actuators():
        """
        Return a list of available actuators in the current installation of PyPads.
        :return: Actuator classes
        """
        from pypads.app.actuators import actuators
        return actuators()

    @staticmethod
    def available_cmds():
        """
        Return a list of available api commands in the current installation of PyPads.
        :return: API commands classes
        """
        from pypads.app.api import api
        return api()

    @property
    def cache(self) -> PypadsCache:
        """
        Return the cache of pypads. This holds generic cache data and run cache data.
        :return:
        """
        return self._cache

    @property
    def uri(self):
        """
        Return the tracking uri pypads uses for mlflow.
        :return:
        """
        return self._cache.get("uri")

    @property
    def folder(self):
        """
        Return the local folder pypads uses for its temporary storage, configuration etc.
        :return:
        """
        return self._cache.get("folder")

    @property
    def config(self):
        """
        Return the configuration of pypads.
        :return: Configuration dict
        """
        if self._cache.exists("config"):
            return self._cache.get("config")
        if self.api.active_run() is not None:
            tags = self.mlf.get_run(mlflow.active_run().info.run_id).data.tags
            if CONFIG_NAME not in tags:
                raise Exception("Config for pypads is not defined.")
            try:
                return ast.literal_eval(tags[CONFIG_NAME])
            except Exception as e:
                raise Exception("Config for pypads is malformed. " + str(e))
        else:
            return {}

    @config.setter
    def config(self, value: dict):
        """
        Set the configuration of pypads.
        :param value: Dict containing config key, values
        :return:
        """
        # Set the config as tag
        if self.api.active_run() is not None:
            mlflow.set_tag(CONFIG_NAME, value)
        else:
            # Set the config as soon as the run is started as tag
            def set_config(*args, **kwargs):
                mlflow.set_tag(CONFIG_NAME, value)

            self.api.register_setup_fn("config_persist", "Function persisting the current pypads configuration.",
                                       set_config, nested=False, intermediate=False)
        self._cache.add("config", value)

    @property
    def mapping_registry(self):
        """
        Access the mapping registry holding all references to mapping files etc.
        :return: Mapping registry
        """
        return self._mapping_registry

    @property
    def mappings(self):
        """
        Get the to the constructor passed mappings.
        :return: Additional mapping files.
        """
        return self._cache.get("mappings")

    @property
    def hook_registry(self):
        """
        Access the hook registry holding all mappings of anchors/hooks to events
        :return: Hook registry
        """
        return self._hook_registry

    @property
    def hooks(self):
        """
        Get the to the constructor passed hook / event mappings.
        :return: Hook configuration.
        """
        return self._cache.get("hooks")

    @property
    def function_registry(self):
        """
        Access the function registry holding all mappings of events to logging functions
        :return: Function registry
        """
        return self._function_registry

    @property
    def events(self):
        """
        Get the to the constructor passed event / logging function mappings.
        :return: Hook configuration.
        """
        return self._cache.get("events")

    @property
    def api(self):
        """
        Access the api of pypads.
        :return: PyPadsAPI
        """
        return self._api

    @property
    def decorators(self):
        """
        Access the decorators of pypads.
        :return: PyPadsDecorators
        """
        return self._decorators

    @property
    def validators(self):
        """
        Access the validators of pypads.
        :return: PyPadsValidators
        """
        return self._validators

    @property
    def actuators(self):
        """
        Access the actuators of pypads.
        :return: PyPadsActuators
        """
        return self._actuators

    @property
    def wrap_manager(self):
        """
        Return the wrap manager of PyPads. This is used to wrap modules, functions and classes.
        :return: WrapManager
        """
        return self._wrap_manager

    @property
    def call_tracker(self):
        """
        Return the call tracker of PyPads. This is used to keep track of the calls of hooked functions.
        :return: CallTracker
        """
        return self._call_tracker

    @property
    def managed_git_factory(self):
        """
        Return Git manager of PyPads. This used to create and manages git repositories.
        :return: ManagedGitFactory
        """
        return self._managed_git_factory

    @property
    def backend(self):
        """
        Return the backend of PyPads. This is currently hardcoded to mlflow.
        :return: Backend
        """
        return self._backend

    @property
    def mlf(self):
        """
        Return a mlflow client to interact with stored data.
        :return: MlflowClient
        """
        return self._backend.mlf

    def add_atexit_fn(self, fn):
        """
        Add function to be executed before stopping your process. This function is also added to pypads and errors
        are caught to not impact the experiment itself. Deactivating pypads should be able to run some of the atExit fns
        declared for pypads.
        """

        def defensive_atexit():
            try:
                return fn()
            except (KeyboardInterrupt, Exception) as e:
                logger.error("Couldn't run atexit function " + fn.__name__ + " because of " + str(e))

        self._atexit_fns.append(defensive_atexit)
        atexit.register(defensive_atexit)

    def is_affected_module(self, name, affected_modules=None):
        """
        Check if a given module name is in the list of affected modules.
        You can pass a list of affected modules or take the one of the wrap_manager.
        :param name: Name of the module to check
        :param affected_modules: The list of affected modules (can be None and will be extracted automatically)
        :return:
        """
        if affected_modules is None:
            affected_modules = self.wrap_manager.module_wrapper.punched_module_names

        affected = set([module.split(".", 1)[0] for module in affected_modules])
        return any([name.startswith(module + ".") or name == module for module in affected]) or any(
            [name.startswith(module + ".") or name == module for module in
             affected_modules])

    def activate_tracking(self, reload_modules=False, reload_warnings=True, clear_imports=False, affected_modules=None):
        """
        Function to duck punch all objects defined in the mapping files. This should at best be called before importing
        any libraries.
        :param affected_modules: Affected modules of the mapping files.
        :param clear_imports: Clear imports after punching. CAREFUL THIS IS EXPERIMENTAL!
        :param reload_warnings: Show warnings of affected modules which were already imported before the importlib was extended.
        :param reload_modules: Force a reload of affected modules. CAREFUL THIS IS EXPERIMENTAL!
        :return:
        """
        if affected_modules is None:
            # Modules are affected if they are mapped by a library or are already punched
            affected_modules = self.wrap_manager.module_wrapper.punched_module_names | \
                               {l.name for l in self.mapping_registry.get_libraries()}

        global tracking_active
        if not tracking_active:
            from pypads.app.pypads import set_current_pads
            set_current_pads(self)

            # Add our loader to the meta_path
            extend_import_module()

            import sys
            import importlib
            loaded_modules = [(name, module) for name, module in sys.modules.items()]
            for name, module in loaded_modules:
                if self.is_affected_module(name, affected_modules):
                    if reload_warnings:
                        logger.warning(
                            name + " was imported before PyPads. To enable tracking import PyPads before or use "
                                   "reload_modules / clear_imports. Every already created instance is not tracked.")

                    if clear_imports:
                        del sys.modules[name]

                    if reload_modules:
                        try:
                            spec = importlib.util.find_spec(module.__name__)
                            duck_punch_loader(spec)
                            loader = spec.loader
                            module = loader.load_module(module.__name__)
                            loader.exec_module(module)
                            importlib.reload(module)
                        except Exception as e:
                            logger.debug("Couldn't reload module " + str(e))

            tracking_active = True
        else:
            raise Exception("Currently only one tracker can be activated at once.")
        return self

    def deactivate_tracking(self, run_atexits=False, reload_modules=True):
        """
        Deacticate the current tracking and cleanup.
        :param run_atexits: Run the registered atexit functions of pypads
        :param reload_modules: Force a reload of affected modules
        :return:
        """
        # run atexit fns if needed
        if run_atexits:
            for fn in self._atexit_fns:
                fn()

        # Remove atexit fns
        for fn in self._atexit_fns:
            atexit.unregister(fn)

        import sys
        import importlib
        loaded_modules = [(name, module) for name, module in sys.modules.items()]
        for name, module in loaded_modules:
            if self.is_affected_module(name):
                del sys.modules[name]

                if reload_modules:
                    # reload modules if they where affected
                    try:
                        spec = importlib.util.find_spec(module.__name__)
                        duck_punch_loader(spec)
                        loader = spec.loader
                        module = loader.load_module(module.__name__)
                        loader.exec_module(module)
                        importlib.reload(module)
                    except Exception as e:
                        logger.debug("Couldn't reload module " + str(e))

        global tracking_active
        tracking_active = False
        # noinspection PyTypeChecker
        from pypads.app.pypads import set_current_pads
        set_current_pads(None)

    def start_track(self, experiment_name=None, disable_run_init=False):
        """
        Start a new run to track.
        :param experiment_name: The name of the mlflow experiment
        :param disable_run_init: Flag to indicate if the run_init functions are to be run on an already existing run.
        :return:
        """
        if not tracking_active:
            self.activate_tracking()

        # check if there is already an active run
        run = mlflow.active_run()
        if run is None:
            experiment_name = experiment_name or DEFAULT_EXPERIMENT_NAME
            # Create run if run doesn't already exist
            experiment = mlflow.get_experiment_by_name(experiment_name)
            experiment_id = experiment.experiment_id if experiment else mlflow.create_experiment(experiment_name)
            run = self.api.start_run(experiment_id=experiment_id)
        else:
            # Run init functions if run already exists but tracking is starting for it now
            if not disable_run_init:
                self.api.run_setups()

        experiment = self.backend.mlf.get_experiment_by_name(
            experiment_name) if experiment_name else self.backend.mlf.get_experiment(run.info.experiment_id)

        # override active run if used
        if experiment_name and run.info.experiment_id is not experiment.experiment_id:
            logger.warning("Active run doesn't match given input name " + experiment_name + ". Recreating new run.")
            try:
                self.api.start_run(experiment_id=experiment_name, nested=True)
            except Exception:
                mlflow.end_run()
                self.api.start_run(experiment_id=experiment_name)
        return self


# --- Pypads Plugins ---
discovered_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg
    in pkgutil.iter_modules()
    if name.startswith('pypads_')
}



import inspect
from _py_abc import ABCMeta
from abc import abstractmethod
from copy import copy
from types import ModuleType
from typing import Set, Type

from pydantic import BaseModel

from pypads import logger
from pypads.importext.mappings import MatchedMapping
from pypads.model.metadata import ModelHolder
from pypads.model.models import ContextModel


def fullname(o):
    """
    Build the full name for a given object
    :param o: object
    :return:
    """
    if isinstance(o, ModuleType):
        return o.__name__
    module = o.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Don't report __builtin__
    else:
        return module + '.' + o.__name__


class Context(ModelHolder):
    """
    Context of the wrapping. In general this is a class or module
    """

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return ContextModel

    def __init__(self, context, reference=None, *args, **kwargs):
        if context is None:
            raise ValueError("A context has to be passed for a object to be wrapped.")
        reference = reference if reference is not None else fullname(context)
        super().__init__(*args, reference=reference, **kwargs)
        self._c = context

    def overwrite(self, key, obj):
        setattr(self._c, key, obj)

    def has_wrap_meta(self, mapping, wrappee):
        if not inspect.isfunction(wrappee):
            holder = wrappee
        else:
            holder = self._c
        # for k in holder.__dict__:
        #     if k.startswith("_pypads_mapping_"):
        #         if mapping in getattr(holder, k):
        #             return True
        if hasattr(holder, "_pypads_mapping_" + wrappee.__name__):
            for hit in getattr(holder, "_pypads_mapping_" + wrappee.__name__):
                if hit.mapping == mapping:
                    return True
        return False

    def store_wrap_meta(self, matched_mapping: MatchedMapping, wrappee):
        try:
            if not inspect.isfunction(wrappee) or "<slot wrapper" in str(wrappee):
                holder = wrappee
            else:
                holder = self._c
            # Set self reference
            if not hasattr(holder, "_pypads_mapping_" + wrappee.__name__):
                setattr(holder, "_pypads_mapping_" + wrappee.__name__, set())
            getattr(holder, "_pypads_mapping_" + wrappee.__name__).add(matched_mapping)

        except TypeError as e:
            logger.debug("Can't set attribute '" + wrappee.__name__ + "' on '" + str(self._c) + "'.")
            raise e

    def store_original(self, wrappee):
        try:
            if not inspect.isfunction(wrappee):
                holder = wrappee
            else:
                holder = self._c
            setattr(holder, self.original_name(wrappee), copy(wrappee))
        except TypeError as e:
            logger.debug("Can't set attribute '" + wrappee.__name__ + "' on '" + str(self._c) + "'.")
            return self._c

    def has_original(self, wrappee):
        return hasattr(self._c, self.original_name(wrappee)) or hasattr(wrappee,
                                                                        self.original_name(wrappee))

    def defined_stored_original(self, wrappee):
        if not inspect.isfunction(wrappee):
            return self.original_name(wrappee) in wrappee.__dict__
        else:
            return self.original_name(wrappee) in self._c.__dict__

    def original_name(self, wrappee):
        return "_pypads_original_" + str(wrappee.__name__)

    def original(self, wrappee):
        if not inspect.isfunction(wrappee) and not inspect.ismethod(wrappee):
            try:
                return getattr(wrappee, self.original_name(wrappee))
            except AttributeError:
                for attr in dir(wrappee):
                    if attr.endswith("_" + wrappee.__name__) and attr.startswith("_pypads_original_"):
                        return getattr(wrappee, attr)
        else:
            try:
                return getattr(self._c, self.original_name(wrappee))
            except AttributeError:
                for attr in dir(self._c):
                    if attr.endswith("_" + wrappee.__name__) and attr.startswith("_pypads_original_"):
                        return getattr(self._c, attr)

    def is_class(self):
        return inspect.isclass(self._c)

    def is_module(self):
        return inspect.ismodule(self._c)

    def real_context(self, fn_name):
        """
        Find where the function was defined
        :return:
        """

        # If the context is not an class it has to define the function itself
        if not self.is_class():
            if hasattr(self._c, fn_name):
                return self
            else:
                logger.warning("Context " + str(self._c) + " of type " + type(
                    self._c) + " doesn't define " + fn_name)
                return None

        # Find defining class by looking at the __dict__ and mro
        defining_class = None
        try:
            mro = self._c.mro()
            for clazz in mro[0:]:
                defining_class = clazz
                if hasattr(clazz, "__dict__") and fn_name in defining_class.__dict__:
                    if callable(defining_class.__dict__[fn_name]):
                        break
                    else:
                        # TODO do we need a workaround for
                        #  <sklearn.utils.metaestimators._IffHasAttrDescriptor object at 0x121e56810> again?
                        break
        except Exception as e:
            logger.warning("Couldn't get defining class of context '" + str(
                self._c) + ".")
            return self._c

        if defining_class and defining_class is not object:
            return Context(defining_class)
        return None

    @property
    def container(self):
        return self._c

    def get_dict(self):
        return self._c.__dict__

    def __str__(self):
        return str(self._c)


class BaseWrapper:
    __metaclass__ = ABCMeta

    def __init__(self, pypads):
        from pypads.app.base import PyPads
        self._pypads: PyPads = pypads

    @abstractmethod
    def wrap(self, wrappee, ctx, matched_mappings: Set[MatchedMapping]):
        raise NotImplementedError()

    def _get_hooked_fns(self, matched_mappings: Set[MatchedMapping]):
        """
        For a given fn find the hook functions defined in a mapping and configured in a configuration.
        :param fn:
        :param mapping:
        :return:
        """
        fns = []
        hooks = set()
        for matched_mapping in matched_mappings:
            for hook in matched_mapping.mapping.hooks:
                hooks.add(hook)

        for hook in hooks:
            fns = fns + self._pypads.hook_registry.get_logging_functions(hook)
        return fns

    @classmethod
    def _get_current_config(cls):
        from pypads.app.pypads import get_current_config
        return get_current_config(default={"events": {}, "recursive": True})



import os
from typing import List, Type

from pydantic import HttpUrl, BaseModel

from app.env import LoggerEnv
from pypads import logger
from pypads.app.injections.base_logger import LoggerCall, TrackedObject
from pypads.app.injections.run_loggers import RunSetup
from pypads.model.models import TrackedObjectModel, LibraryModel, OutputModel, ArtifactMetaModel
from pypads.utils.logging_util import WriteFormats


class DependencyTO(TrackedObject):
    """
    Tracking object class for run env info, i.e dependencies.
    """

    class DependencyModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/env/Dependencies"

        dependencies: List[LibraryModel] = []
        content_format: WriteFormats = WriteFormats.text

        class Config:
            orm_mode = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.DependencyModel

    def __init__(self, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)

    def _add_dependency(self, pip_freeze):
        for item in pip_freeze:
            name, version = item.split('==')
            self.dependencies.append(LibraryModel(name=name, version=version))
        path = os.path.join(self._base_path(), self._get_artifact_path("pip_freeze"))
        self._store_artifact("\n".join(pip_freeze),
                             ArtifactMetaModel(path=path, description="dependency list from pip freeze",
                                               format=WriteFormats.text))

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), "Env", name)


class DependencyRSF(RunSetup):
    """Store information about dependencies used in the experimental environment."""

    name = "Dependencies Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/dependency-run-logger"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    _dependencies = {"pip"}

    class DependencyRSFOutput(OutputModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/DependencyRSF-Output"

        dependencies: DependencyTO.get_model_cls() = None

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.DependencyRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        pads = _pypads_env.pypads
        logger.info("Tracking execution to run with id " + pads.api.active_run().info.run_id)
        dependencies = DependencyTO(tracked_by=_logger_call)
        try:
            # Execute pip freeze
            try:
                # noinspection PyProtectedMember,PyPackageRequirements
                from pip._internal.operations import freeze
            except ImportError:  # pip < 10.0
                # noinspection PyUnresolvedReferences,PyPackageRequirements
                from pip.operations import freeze
            dependencies._add_dependency(list(freeze.freeze()))
        except Exception as e:
            _logger_output.set_failure_state(e)
        finally:
            dependencies.store(_logger_output, "dependencies")


class LoguruTO(TrackedObject):
    """
    Tracking object class for run env info, i.e dependencies.
    """

    class LoguruModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/env/Logs"

        meta: ArtifactMetaModel = ...

        class Config:
            orm_mode = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.LoguruModel

    def __init__(self, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)
        path = os.path.join(self._base_path(), self._get_artifact_path("logs.log"))
        self.meta = ArtifactMetaModel(path=path, description="Logs of the current run", format=WriteFormats.text)


class LoguruRSF(RunSetup):
    """Store all logs of the current run into a file."""

    name = "Loguru Run Setup Logger"
    uri = "https://www.padre-lab.eu/onto/loguru-run-logger"

    _dependencies = {"loguru"}

    class LoguruRSFOutput(OutputModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/LoguruRSF-Output"

        logs: LoguruTO.get_model_cls() = ...

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.LoguruRSFOutput

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        pads = _pypads_env.pypads

        from pypads.app.api import PyPadsApi
        _api: PyPadsApi = pads.api

        from pypads.utils.logging_util import get_temp_folder
        folder = get_temp_folder()

        logs = LoguruTO(tracked_by=_logger_call)

        # TODO loguru has problems with multiprocessing / make rotation configurable etc
        from pypads.pads_loguru import logger_manager
        lid = logger_manager.add(os.path.join(folder, "run_" + _api.active_run().info.run_id + ".log"),
                                 rotation="50 MB",
                                 enqueue=True)

        import glob

        def remove_logger(pads, *args, **kwargs):
            try:
                from pypads.pads_loguru import logger_manager
                logger_manager.remove(lid)
            except Exception:
                pass
            for file in glob.glob(os.path.join(folder, "run_*.log")):
                pads.api.log_artifact(file, artifact_path=logs.meta.path)

        logs.store(_logger_output, "logs")
        _api.register_teardown_fn("logger_" + str(lid), remove_logger)



class Plugin:
    """
    A plugin holding instances of a given type of extension elements.
    """

    def __init__(self, type):
        self._type = type

    @property
    def type(self):
        return self._type

    @property
    def _get_meta(self):
        """
        Todo return meta (for example where plugin was defined)
        :return:
        """
        return self.__module__

    def __hash__(self):
        return hash(self.__class__)

    def __eq__(self, other):
        """
        Plugins are to be considered equal if they are of the same class. TODO Singleton?
        :param other:
        :return:
        """
        return self.__class__ == other.__class__


class ExtendableMixin:
    """
    Abstract class for plugin managers.
    """

    def __init__(self, plugin_list=None):
        """
        Abstract class for plugin managers.
        :param instances: Initial plugins which should be added.
        """
        # if instances is not None:
        #     if not isinstance(instances, Iterable):
        #         instances = [instances]
        #     self._instances = instances
        # else:
        self._instances = plugin_list

    def __getattr__(self, item):
        """ Try to find a fitting variable / function. Older definitions overwrite newer. """
        if item is "__get_instance":
            return self.__get_instance

        for i in self._instances:
            if hasattr(i, item):
                return getattr(i, item)

    def __get_instance(self, index):
        return self._instances[index]



import traceback
from abc import abstractmethod, ABCMeta
from typing import List, Union, Tuple, Set

from pypads import logger
from pypads.app.misc.inheritance import SuperStop
from pypads.importext.versioning import LibSelector, VersionNotFoundException
from pypads.model.metadata import ModelObject
from pypads.model.models import LibraryModel

DEFAULT_ORDER = 1


class NoCallAllowedError(Exception):
    """
    Exception to denote that a callable couldn't be called, but isn't essential.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MissingDependencyError(NoCallAllowedError):
    """
    Exception to be thrown if a dependency is missing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class OrderMixin(SuperStop):
    """
    Object defining an order attribute to denote its priority.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, order=DEFAULT_ORDER, **kwargs):
        self._order = order
        super().__init__(*args, **kwargs)

    @property
    def order(self):
        return self._order

    @staticmethod
    def sort(collection, reverse=False):  # type: (List[OrderMixin]) -> List[OrderMixin]
        copy = collection.copy()
        copy.sort(key=lambda e: e.order, reverse=reverse)
        return copy

    @staticmethod
    def sort_mutable(collection, reverse=False):  # type: (List[OrderMixin]) -> None
        collection.sort(key=lambda e: e.order, reverse=reverse)


class CallableMixin(SuperStop):
    """
    Object defining a _call method which can be overwritten.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.__real_call__(*args, **kwargs)

    @abstractmethod
    def __real_call__(self, *args, **kwargs):
        pass


class DependencyMixin(CallableMixin):
    """
    Callable being able to be disabled / enabled depending on package availability.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    """
    Overwrite this to provide your package names.
    :return: List of needed packages by the logger.
    """
    _dependencies: Set[Union[LibSelector, str, Tuple[str, str]]] = set()

    @property
    def dependencies(self):
        return self._to_lib_selectors(self._dependencies)

    @staticmethod
    def _to_lib_selectors(dependencies: Set[Union[LibSelector, str, Tuple[str, str]]]) -> Set[LibSelector]:
        selectors = set()
        for d in dependencies:
            selectors.add(LibSelector(name=d[0], constraint=d[1]) if isinstance(d, tuple) else LibSelector(
                name=d) if not isinstance(d, LibSelector) else d)
        return selectors

    def _check_dependencies(self):
        """
        Raise error if dependencies are missing.
        """
        missing = []
        selectors = self.dependencies
        if selectors is not None:
            for selector in selectors:
                try:
                    if not selector.is_installed():
                        missing.append(selector)
                except VersionNotFoundException as e:
                    # TODO couldn't get version allow execution for now
                    pass
        if len(missing) > 0:
            raise MissingDependencyError(
                "Can't log " + str(self) + ". Missing dependencies: " + ", ".join([str(d) for d in missing]))

    def __call__(self, *args, **kwargs):
        self._check_dependencies()
        return super().__call__(*args, **kwargs)


class IntermediateCallableMixin(CallableMixin):
    """
    Callable being able to be disable / enabled on nested / intermediate runs.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, nested=True, intermediate=True, **kwargs):
        self._intermediate = intermediate
        self._nested = nested
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        from pypads.app.pypads import is_nested_run
        if self._nested or not is_nested_run():
            from pypads.app.pypads import is_intermediate_run
            if self._intermediate or not is_intermediate_run():
                return super().__call__(*args, **kwargs)
        raise NoCallAllowedError("Call wasn't allowed by intermediate / nested settings of the current run.")

    @property
    def allow_nested(self):
        return self._nested

    @property
    def allow_intermediate(self):
        return self._intermediate


class TimedCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable tracking its own execution time.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        c = super().__call__
        from pypads.injections.analysis.time_keeper import timed
        _return, time = timed(lambda: c(*args, **kwargs))
        return _return, time


class DefensiveCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable handling errors produced by itself.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, ctx, *args, _pypads_env=None, **kwargs):
        try:
            return super().__call__(ctx, *args, _pypads_env=_pypads_env, **kwargs)
        except KeyboardInterrupt:
            return self._handle_error(*args, ctx=ctx, _pypads_env=_pypads_env, error=Exception("KeyboardInterrupt"),
                                      **kwargs)
        except Exception as e:
            import traceback
            logger.debug(traceback.format_exc())
            return self._handle_error(*args, ctx=ctx, _pypads_env=_pypads_env, error=e, **kwargs)

    @abstractmethod
    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        raise NotImplementedError()


class ConfigurableCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable storing additional creation args as fields to be accessible later on.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        self._kwargs = kwargs
        super().__init__(*args, **kwargs)

    @property
    def static_parameters(self):
        return self._kwargs


class LibrarySpecificMixin(SuperStop):
    """
    A class only being applicable for a certain library.
    """
    __metaclass__ = ABCMeta

    supported_libraries: Set[LibSelector] = set()

    def allows_any(self, lib_selector: LibSelector):
        libraries = self.supported_libraries
        return len(libraries) == 0 or any([s.allows_any(lib_selector) for s in libraries])

    def allows(self, version):
        libraries = self.supported_libraries
        return len(libraries) == 0 or any([s.allows(version) for s in libraries])

    def is_applicable(self, lib_selector: LibSelector, only_name=True):
        if self.allows_any(lib_selector):
            return True
        if only_name:
            for s in self.supported_libraries:
                if s.name == lib_selector.name:
                    return True
        return False


class FunctionHolderMixin(CallableMixin):
    """
    Holds the given function in a timed callable.
    """

    def __init__(self, *args, fn, **kwargs):
        super().__init__(*args, **kwargs)
        self._fn = fn

    @property
    def fn(self):
        return self._fn

    def __real_call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


class ProvenanceMixin(ModelObject, metaclass=ABCMeta):
    """
    Class extracting its library reference automatically if possible.
    """

    def __init__(self, *args, lib_model: LibraryModel = None, **kwargs):
        super().__init__(*args, **kwargs)
        if lib_model is None:
            setattr(self, "defined_in", self._get_library_descriptor())
        else:
            setattr(self, "defined_in", lib_model)

        if not hasattr(self, "uri") or getattr(self, "uri") is None:
            setattr(self, "uri", "{}#{}".format(getattr(self, "is_a"), self.uid))

    def _get_library_descriptor(self) -> LibraryModel:
        """
        Try to extract the defining package of this class.
        :return:
        """
        # TODO extract reference to self package
        try:
            name = self.__module__.split(".")[0]
            from pypads.utils.util import find_package_version
            version = find_package_version(name)
            return LibraryModel(name=name, version=version, extracted=True)
        except Exception:
            return LibraryModel(name="__unkown__", version="0.0", extracted=True)


class BaseDefensiveCallableMixin(DefensiveCallableMixin):
    """
    Defensive callable ignoring errors but printing a warning to console.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, error_message=None, **kwargs):
        self._message = error_message if error_message else "Couldn't execute {}, because of exception: {} \nTrace:\n{}"
        super().__init__(*args, **kwargs)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        logger.warning(self._message.format("{}.{}".format(self.__class__.__name__, self.__name__), str(error),
                                            traceback.format_exc()))



import os
import traceback
from abc import abstractmethod, ABCMeta
from typing import Type, Set, List, Callable

import mlflow
from pydantic import HttpUrl, BaseModel

from app.env import LoggerEnv
from pypads import logger
from pypads.app.misc.mixins import DependencyMixin, DefensiveCallableMixin, TimedCallableMixin, \
    IntermediateCallableMixin, NoCallAllowedError, ConfigurableCallableMixin, LibrarySpecificMixin, \
    FunctionHolderMixin, ProvenanceMixin, BaseDefensiveCallableMixin
from pypads.importext.versioning import LibSelector
from pypads.injections.analysis.time_keeper import TimingDefined
from pypads.model.models import MetricMetaModel, \
    ParameterMetaModel, ArtifactMetaModel, TrackedObjectModel, LoggerCallModel, OutputModel, EmptyOutput, TagMetaModel
from pypads.utils.logging_util import WriteFormats


class PassThroughException(Exception):
    """
    Exception to be passed from _pre / _post and not be caught by the defensive logger.
    """

    def __init__(self, *args):
        super().__init__(*args)


class OriginalExecutor(FunctionHolderMixin, TimedCallableMixin):
    """
    Class adding a time tracking to the original execution given as fn.
    """

    def __init__(self, *args, fn, **kwargs):
        super().__init__(*args, fn=fn, **kwargs)


class LoggerExecutor(DefensiveCallableMixin, FunctionHolderMixin, TimedCallableMixin, ConfigurableCallableMixin):
    __metaclass__ = ABCMeta
    """
    Executor for the functionality a tracking function provides.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        """
        Function to handle an error executing the logging functionality. In general this should add a failure tag and
        log to console.
        :param args: Arguments passed to the function
        :param ctx: Context of the function
        :param _pypads_env: Pypads environment
        :param error: Exception which was raised on the execution
        :param kwargs: Kwargs passed to the function
        :return:
        """
        try:
            raise error
        except TimingDefined:

            # Ignore if due to timing defined
            pass
        except NotImplementedError:

            # Ignore if only pre or post where defined
            return None, 0
        except (NoCallAllowedError, PassThroughException) as e:

            # Pass No Call Allowed Error through
            raise e
        except Exception as e:

            # Catch other exceptions for this single logger
            try:
                mlflow.set_tag("pypads_failure", str(error))
            except Exception as e:
                pass
            logger.error(
                f"Tracking failed for {str(_pypads_env)} with: {str(error)} \nTrace:\n{traceback.format_exc()}")
            return None, 0


class LoggerCall(ProvenanceMixin):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return LoggerCallModel

    def __init__(self, *args, logging_env: LoggerEnv, **kwargs):
        super().__init__(*args, **kwargs)
        self._logging_env = logging_env

    def store(self):
        from pypads.app.pypads import get_current_pads
        from pypads.utils.logging_util import WriteFormats
        get_current_pads().api.log_mem_artifact("{}".format(str(self.uid)), self.json(), WriteFormats.json.value,
                                                path=self.created_by + "Calls")


class TrackedObject(ProvenanceMixin):
    is_a = "https://www.padre-lab.eu/onto/tracked_object"

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return TrackedObjectModel

    def __init__(self, *args, tracked_by, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)

    @staticmethod
    def _store_metric(val, meta: MetricMetaModel):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        consolidated_json = pads.cache.get('consolidated_dict', None)
        if consolidated_json is not None:
            metrics_dict = consolidated_json.get('metrics', {})
            metrics_list = metrics_dict.get(meta.name, [])
            metrics_list.append(val)
            metrics_dict[meta.name] = metrics_list
            consolidated_json['metrics'] = metrics_dict
            pads.cache.add('consolidated_dict', consolidated_json)
        pads.api.log_metric(meta.name, val, meta=meta)

    @staticmethod
    def _store_param(val, meta: ParameterMetaModel):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        consolidated_json = pads.cache.get('consolidated_dict', None)
        if consolidated_json is not None:
            # Set the parameter
            estimator_name = meta.name[:meta.name.rfind('.')]
            parameters = consolidated_json.get('parameters', {})
            estimator_dict = parameters.get(estimator_name, {})
            estimator_dict[meta.name.split(sep='.')[-1]] = val

            # Store the dictionaries back into the cache
            parameters[estimator_name] = estimator_dict
            consolidated_json['parameters'] = parameters
            pads.cache.add('consolidated_dict', consolidated_json)

        get_current_pads().api.log_param(meta.name, val, meta=meta)

    @staticmethod
    def _store_artifact(val, meta: ArtifactMetaModel):
        from pypads.app.pypads import get_current_pads
        get_current_pads().api.log_mem_artifact(meta.path, val, meta=meta, write_format=meta.format)

    @staticmethod
    def _store_tag(val, meta: TagMetaModel):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        consolidated_json = pads.cache.get('consolidated_dict', None)
        if consolidated_json is not None:
            tags = consolidated_json.get('tags', dict())
            tags[meta.name] = val
            consolidated_json['tags'] = tags
            pads.cache.add('consolidated_json', consolidated_json)
        pads.api.set_tag(meta.name, val)

    def _base_path(self):
        return os.path.join(self.tracked_by.created_by, "TrackedObjects", self.__class__.__name__)

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)),name)

    def store(self, output, key="tracked_object", *json_path):
        """
        :param output:
        :param key: Name of the tracking object in the schema
        :param json_path: path in the output schema
        :return:
        """
        output.add_tracked_object(self, key, *json_path)


class LoggerOutput(ProvenanceMixin):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return OutputModel

    def add_tracked_object(self, to: TrackedObject, key, *json_path):
        curr = self
        for p in json_path:
            curr = getattr(curr, p)
        if hasattr(curr, key):
            attr = getattr(curr, key)
            if isinstance(attr, List):
                attr.append(to)
                to = attr
            elif isinstance(attr, Set):
                attr.add(to)
                to = attr
        setattr(curr, key, to)

    def store(self, path=""):
        from pypads.app.pypads import get_current_pads
        return get_current_pads().api.store_logger_output(self, path)

    def set_failure_state(self, e: Exception):
        self.failed = "Logger Output might be inaccurate/corrupt due to exception in execution: '{}'".format(str(e))


class Logger(BaseDefensiveCallableMixin, IntermediateCallableMixin, DependencyMixin,
             LibrarySpecificMixin, ProvenanceMixin, ConfigurableCallableMixin, metaclass=ABCMeta):
    """
    Generic tracking function used for storing information to a backend.
    """

    is_a: HttpUrl = "https://www.padre-lab.eu/onto/tracking-function"

    # Default allow all libraries
    supported_libraries = {LibSelector(name=".*", constraint="*")}
    _schema_path = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracked_objects: Set[TrackedObject] = set()
        self._cleanup_fns = {}

    @classmethod
    def build_output(cls, **kwargs):
        schema_class = cls.output_schema_class()

        class OutputModelHolder(LoggerOutput):

            @classmethod
            def get_model_cls(cls) -> Type[BaseModel]:
                return schema_class

        return OutputModelHolder(**kwargs)

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return EmptyOutput

    @classmethod
    def output_schema(cls):
        return cls.output_schema_class().schema()

    @classmethod
    def _default_output_class(cls, clazz: Type[TrackedObject]) -> Type[OutputModel]:
        class DefaultOutput(OutputModel):
            tracked_object: clazz.get_model_cls() = ...  # Path to tracking objects

            class Config:
                orm_mode = True

        return DefaultOutput

    @classmethod
    def store_schema(cls, path=None):
        if not cls._schema_path:
            path = path or ""
            from pypads.app.pypads import get_current_pads
            schema_path = os.path.join(path, cls.__name__ + "_schema")
            get_current_pads().api.log_mem_artifact(schema_path,
                                                    cls.schema(), write_format=WriteFormats.json)
            get_current_pads().api.log_mem_artifact(os.path.join(path, cls.__name__ + "_output_schema"),
                                                    cls.output_schema(), write_format=WriteFormats.json)
            cls._schema_path = path
        return cls._schema_path

    @abstractmethod
    def _base_path(self):
        return "Loggers/"

    def cleanup_fns(self, call: LoggerCall) -> List[Callable]:
        return self._cleanup_fns[call] if call in self._cleanup_fns.keys() else []

    def register_cleanup_fn(self, call: LoggerCall, fn):
        if call not in self._cleanup_fns:
            self._cleanup_fns[call] = []
        self._cleanup_fns[call].append(fn)


class SimpleLogger(Logger):

    def __init__(self, *args, fn=None, **kwargs):
        super().__init__(*args, **kwargs)
        if fn is None:
            fn = self._call
        if not hasattr(self, "_fn"):
            self._fn = LoggerExecutor(fn=fn)

    @property
    def __name__(self):
        if self._fn.fn != self._call:
            return self._fn.fn.__name__
        else:
            return self.__class__.__name__

    def _call(self, _pypads_env: LoggerEnv, _logger_call: LoggerCall, _logger_output,*args, **kwargs):
        """
        Function where to add you custom code to execute before starting or ending the run.

        :param pads: the current instance of PyPads.
        """
        pass

    def __real_call__(self, *args, _pypads_env: LoggerEnv, **kwargs):
        self.store_schema(self._base_path())

        _pypads_params = _pypads_env.parameter

        logger_call = self.build_call_object(_pypads_env, created_by=self.store_schema(self._base_path()))
        output = self.build_output()

        try:
            _return, time = self._fn(*args, _pypads_env=_pypads_env, _logger_call=logger_call, _logger_output=output,
                                     _pypads_params=_pypads_params,
                                     **{**self.static_parameters, **kwargs})

            logger_call.execution_time = time
        except Exception as e:
            logger_call.failed = str(e)
            output.set_failure_state(e)
            raise e
        finally:
            for fn in self.cleanup_fns(logger_call):
                fn(self, logger_call)
            logger_call.output = output.store(self._base_path())
            logger_call.store()
        return _return

    @abstractmethod
    def build_call_object(self, _pypads_env, **kwargs):
        raise NotImplementedError()



# import re
# from typing import Sequence
#
# from jsonschema import ValidationError, validators
#
#
# # TODO move defaults to json schema
# # def extend_with_default(validator_class):
# #     # see https://python-jsonschema.readthedocs.io/en/stable/faq/
# #     validate_properties = validator_class.VALIDATORS["properties"]
# #
# #     def set_defaults(validator, properties, instance, schema):
# #         for property, subschema in properties.items():
# #             if "default" in subschema:
# #                 instance.setdefault(property, subschema["default"])
# #
# #         for error in validate_properties(
# #                 validator, properties, instance, schema,
# #         ):
# #             yield error
# #
# #     return validators.extend(
# #         validator_class, {"properties": set_defaults},
# #     )
# from pypads.model.validation import ValidationErrorHandler
#
#
# # def padre_enum(validator, padre_enum, instance, schema):
# #     """
# #     Function to evaluate if a enum exists via jsonschema evaluation. This is used for ontology validation.
# #     :param validator:
# #     :param padre_enum:
# #     :param instance:
# #     :return:
# #     """
# #     if validator.is_type(instance, "string"):
# #         if padre_enum is not None:
# #             # noinspection PyProtectedMember
# #             if not hasattr(PaDREOntology, padre_enum):
# #                 yield ValidationError("%r is not a valid padre enum" % (padre_enum))
# #             # TODO cleanup access to enum
# #             elif instance not in getattr(PaDREOntology, padre_enum)._value2member_map_:
# #                 yield ValidationError("%r is not a valid value entry of padre enum %r" % (instance, padre_enum))
#
#
# padre_schema_validator = validators.extend(validators.Draft7Validator,
#                                            validators={"padre_enum": padre_enum}, version="1.0")
#
#
# class JsonSchemaRequiredHandler(ValidationErrorHandler):
#     """
#     Handle a required error of jsonschema validation.
#     """
#
#     def __init__(self, absolute_path=None, validator=None, get_value=None):
#         super().__init__(absolute_path, validator, get_value)
#
#     @property
#     def validator(self):
#         return self._validator
#
#     @property
#     def absolute_path(self):
#         return self._absolute_path
#
#     def handle(self, obj, e, options):
#         return self.update_options(e, options, super().handle(obj, e, options))
#
#     @staticmethod
#     def update_options(e, options, value):
#         deq = e.absolute_path
#         new = {}
#         current_level = new
#
#         # Build path for the options dict
#         while len(deq) > 0:
#             val = deq.popleft()
#             if len(deq) != 0:
#                 current_level[val] = {}
#                 current_level = current_level[val]
#
#         # Insert value in correct field in options dict
#         if _seq_but_not_str(e.validator_value):
#             for key in e.validator_value:
#
#                 # TODO parsing the jsonschema errors is really ugly there has to be a better way (Own Validator?)
#                 match = re.compile("'(" + re.escape(key) + ")'.*").match(e.message)
#                 if match is not None:
#                     key = match.group(1)
#                     current_level[key] = value
#                     break
#         return {**options, **new}
#
#
# def _seq_but_not_str(obj):
#     return isinstance(obj, Sequence) and not isinstance(obj, (str, bytes, bytearray))



from abc import ABCMeta
from functools import wraps
from typing import Type

from pydantic import HttpUrl, BaseModel

from pypads.app.injections.base_logger import LoggerCall, SimpleLogger
from pypads.app.misc.extensions import ExtendableMixin, Plugin
from pypads.model.models import LoggerModel
from pypads.utils.util import inheritors

actuator_plugins = set()


class Actuator(SimpleLogger, metaclass=ABCMeta):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/actuator"

    def build_call_object(self, _pypads_env, **kwargs):
        return LoggerCall(logging_env=_pypads_env,
                          is_a="https://www.padre-lab.eu/onto/ActuatorLoggerCall", **kwargs)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return LoggerModel

    def _base_path(self):
        return "Actuators/{}/".format(self.__class__.__name__)


class IActuators(Plugin):

    def __init__(self):
        super().__init__(type=Actuator)
        actuator_plugins.add(self)

    def _get_meta(self):
        """ Method returning information about where the actuator was defined."""
        return self.__module__

    def _get_methods(self):
        return [method_name for method_name in dir(self) if callable(getattr(object, method_name))]


def actuator(f):
    """
    Decorator used to convert a function to a tracked actuator.
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        # self is an instance of the class
        return Actuator(fn=f)(self, *args, **kwargs)

    return wrapper


class PyPadsActuators(IActuators):
    def __init__(self):
        super().__init__()

    @property
    def pypads(self):
        from pypads.app.pypads import get_current_pads
        return get_current_pads()

    @actuator
    def set_random_seed(self, seed=None):
        # Set seed if needed
        if seed is None:
            import random
            # import sys
            # seed = random.randrange(sys.maxsize)
            # Numpy only allows for a max value of 2**32 - 1
            seed = random.randrange(2 ** 32 - 1)
        self.pypads.cache.run_add('seed', seed)

        from pypads.injections.analysis.randomness import set_random_seed
        set_random_seed(seed)


class ActuatorPluginManager(ExtendableMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(plugin_list=actuator_plugins)


pypads_actuators = PyPadsActuators()


def actuators():
    """
    Returns classes of
    :return:
    """
    return inheritors(Actuator)



import os
from typing import List, Type

from pydantic import BaseModel, HttpUrl

from pypads.app.injections.base_logger import LoggerCall, TrackedObject
from pypads.app.injections.injection import InjectionLogger
from pypads.model.models import ArtifactMetaModel, TrackedObjectModel, OutputModel
from pypads.utils.logging_util import WriteFormats


# TODO Literal for python 3.7 / 3.8?
class InputTO(TrackedObject):
    """
    Tracking object class for inputs of your tracked workflow.
    """

    class InputModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/FunctionInput"

        class ParamModel(BaseModel):
            content_format: WriteFormats = WriteFormats.pickle
            name: str = ...
            value: ArtifactMetaModel = ...  # path to the artifact containing the param
            type: str = ...

            class Config:
                orm_mode = True
                arbitrary_types_allowed = True

        inputs: List[ParamModel] = []

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.InputModel

    def add_arg(self, name, value, format):
        self._add_param(name, value, format, "argument")

    def add_kwarg(self, name, value, format):
        self._add_param(name, value, format, "keyword-argument")

    def _add_param(self, name, value, format, type):
        path = os.path.join(self._base_path(), self._get_artifact_path(name))
        meta = ArtifactMetaModel(path=path,
                                 description="Input to function with index {} and type {}".format(len(self.inputs),
                                                                                                  type),
                                 format=format)
        self.inputs.append(self.InputModel.ParamModel(content_format=format, name=name, value=meta, type=type))
        self._store_artifact(value, meta)

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), "input", name)


class InputILF(InjectionLogger):
    """
    Function logging the input parameters of the current pipeline object function call.
    """

    name = "InputLogger"
    uri = "https://www.padre-lab.eu/onto/input-logger"

    class InputILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/InputILF-Output"
        FunctionInput: InputTO.get_model_cls() = ...

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls):
        return cls.InputILFOutput

    def __pre__(self, ctx, *args, _pypads_write_format=None, _logger_call: LoggerCall, _logger_output, _args, _kwargs,
                **kwargs):
        """
        :param ctx:
        :param args:
        :param _pypads_write_format:
        :param kwargs:
        :return:
        """

        inputs = InputTO(tracked_by=_logger_call)
        for i in range(len(_args)):
            arg = _args[i]
            inputs.add_arg(str(i), arg, format=_pypads_write_format)

        for (k, v) in _kwargs.items():
            inputs.add_kwarg(str(k), v, format=_pypads_write_format)
        inputs.store(_logger_output,key="FunctionInput")


class OutputTO(TrackedObject):
    """
    Tracking object class for inputs of your tracked workflow.
    """

    def _path_name(self):
        return "inputs"

    class OutputModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/FunctionOutput"

        content_format: WriteFormats = WriteFormats.pickle
        output: str = ...  # Path to the output holding file

        class Config:
            orm_mode = True
            arbitrary_types_allowed = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.OutputModel

    def __init__(self, value, format, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, content_format=format, tracked_by=tracked_by, **kwargs)
        path = os.path.join(self._base_path(), self._get_artifact_path())
        self.output = path
        self._store_artifact(value, ArtifactMetaModel(path=path,
                                                      description="Output of function call {}".format(
                                                          self.tracked_by.original_call),
                                                      format=format))

    def _get_artifact_path(self, name="output"):
        return super()._get_artifact_path(name)


class OutputILF(InjectionLogger):
    """
    Function logging the output of the current pipeline object function call.
    """

    name = "OutputLogger"
    uri = "https://www.padre-lab.eu/onto/output-logger"

    class OutputILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/OutputILF-Output"
        FunctionOutput: OutputTO.get_model_cls() = ...

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls):
        return cls.OutputILFOutput

    def __post__(self, ctx, *args, _pypads_write_format=WriteFormats.pickle, _logger_call, _pypads_result,
                 _logger_output, **kwargs):
        """
        :param ctx:
        :param args:
        :param _pypads_write_format:
        :param kwargs:
        :return:
        """
        output = OutputTO(_pypads_result, format=_pypads_write_format, tracked_by=_logger_call)
        output.store(_logger_output, key="FunctionOutput")



from abc import abstractmethod, ABCMeta
from collections import deque
from typing import Type, List

from pydantic import validate_model, BaseModel, ValidationError

from pypads.app.misc.inheritance import SuperStop
from pypads.model.models import RunObjectModel
from pypads.utils.util import has_direct_attr


class ModelInterface(SuperStop):

    @classmethod
    @abstractmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        raise NotImplementedError("A model cls has to be defined for a class linked to a model.")

    @abstractmethod
    def model(self) -> Type[BaseModel]:
        raise NotImplementedError("A function how to access the model of the class has to be defined.")

    @abstractmethod
    def schema(self):
        raise NotImplementedError("A function how to access the schema of the class has to be defined.")

    @abstractmethod
    def json(self):
        raise NotImplementedError("A function how to convert a class to its json representation has to be defined.")

    @abstractmethod
    def validate(self):
        raise NotImplementedError("A function how to validate the schema for the class has to be defined.")

    def get_model_fields(self):
        """
        Return the field names of a model.
        :return:
        """
        return self.get_model_cls().__fields__


class ModelObject(ModelInterface, metaclass=ABCMeta):
    """
    An object building the model from itself on the fly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        fields = set(self.get_model_fields().keys())

        # Add given fields to metadata object if not already existing
        for key, val in kwargs.items():
            if key in fields and not has_direct_attr(self, key):
                setattr(self, key, val)
                fields.remove(key)

        # Add defaults which are not given
        for key in fields:
            if not has_direct_attr(self, key) and self.get_model_fields():
                setattr(self, key, self.get_model_fields()[key].get_default())
        if issubclass(self.get_model_cls(), RunObjectModel) and (
                not hasattr(self, "uri") or getattr(self, "uri") is None):
            setattr(self, "uri", "{}#{}".format(getattr(self, 'is_a'), getattr(self, 'uid')))

    def model(self):
        return self.get_model_cls().from_orm(self)

    def validate(self):
        validate_model(self.get_model_cls(), self.model().__dict__)

    @classmethod
    def schema(cls):
        schema = cls.get_model_cls().schema()

        # Overwrite the model comment with the comment on the class if one exists
        if cls.__doc__ is not None:
            schema["description"] = cls.__doc__
        return schema

    def json(self, *args, **kwargs):
        return self.model().json(*args, **kwargs)


class ModelHolder(ModelInterface, metaclass=ABCMeta):
    """
    Used for objects storing their information directly into a validated base model
    """

    def __init__(self, *args, model: RunObjectModel = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._model = self.get_model_cls()(**kwargs) if model is None else model

    def __getattr__(self, name):
        if name not in ["_model", "_model_cls"] and name in self.get_model_fields().keys():
            return getattr(self._model, name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name not in ["_model", "_model_cls"] and name in self.get_model_fields().keys():
            setattr(self._model, name, value)
        else:
            return object.__setattr__(self, name, value)

    def model(self):
        return self._model

    def validate(self):
        validate_model(self.get_model_cls(), self._model.__dict__)

    def schema(self):
        return self._model.schema()

    def json(self):
        return self._model.json()


class ModelErrorHandler:
    """ Class to handle errors on the validation of an validatable. """

    def __init__(self, absolute_path=None, validator=None, handle=None):
        self._absolute_path = absolute_path
        self._validator = validator
        self._handle = handle

    @property
    def validator(self):
        return self._validator

    @property
    def absolute_path(self):
        return self._absolute_path

    def handle(self, cls, e, options):
        if (not self._absolute_path or deque(self._absolute_path) == e.absolute_path) and (
                not self._validator or self.validator == e.validator):
            if self._handle is None:
                self._default_handle(e)
            else:
                return self._handle(cls, e, options)
        else:
            raise e

    def _default_handle(self, e):
        print("Empty validation handler triggered: " + str(self))
        raise e


class ModelFactory:

    @staticmethod
    def from_object(cls: BaseModel, obj, handlers: List[ModelErrorHandler] = None):
        return ModelFactory._try_creation(cls, cls.from_orm, handlers=handlers, __history=[], obj=obj)

    @staticmethod
    def make(cls, *args, handlers: List[ModelErrorHandler] = None, **options):
        return ModelFactory._try_creation(cls, cls, *args, handlers=handlers, __history=[], **options)

    @staticmethod
    def _try_creation(cls, fn, *args, handlers: List[ModelErrorHandler] = None, __history=None, **options):
        if handlers is None:
            handlers = []
        try:
            return fn(*args, **options)
        except ValidationError as e:
            # Raise error if we can't handle anything
            if handlers is None:
                raise e
            for handler in handlers:
                new_options = handler.handle(cls, e, options)
                # If the handler could fix the problem return the new value
                if new_options is not None and e not in __history and new_options not in __history:
                    __history.append(e)
                    __history.append(new_options)

                    # Try to create object again
                    return ModelFactory._try_creation(cls, handlers=handlers, __history=__history, **new_options)
            # Raise error if we fail
            raise e



import re
from abc import ABCMeta, abstractmethod

import yaml


class Package:

    def __init__(self, module, package_path):
        self._content = module
        self._path = package_path

    @property
    def content(self):
        return self._content

    @property
    def path(self):  # type: () -> PackagePath
        return self._path

    def __str__(self):
        return str(self._path)


class PackagePathSegment:
    """
    Single segment of a package path. This would be "module", "foo", "bar" in module.foo.bar
    """

    def __init__(self, content: str):
        self._content = content

    @property
    def content(self):
        return self._content

    def __hash__(self):
        return hash(self.content)

    def __eq__(self, other):
        if isinstance(other, str):
            return self._content == other
        return self._content == other.content

    def __str__(self):
        return str(self._content)


class PackagePath:
    """
    Full package path holding all needed segments.
    """

    def __init__(self, content):
        """
        :param content: Takes either a package name like "module.foo.bar" or a list of segments
        """
        if content is None:
            self._segments = []
        else:
            if isinstance(content, str):
                content = self.to_segments(content)
            if not isinstance(content, list):
                raise ValueError("PackagePath has to hold a list of PathSegments.")
            self._segments = content

    @staticmethod
    def to_segments(package):
        """
        Converts given package into segments
        :param package: Package name in form of "module.foo.bar"
        :return: Segments of package name
        """
        return [PackagePathSegment(s) for s in filter(lambda s: len(s) > 0, package.split("."))]

    @property
    def segments(self):
        """
        Returns all segments of the path
        :return:
        """
        return self._segments

    def __eq__(self, other):
        return len(self._segments) is len(other.segments) and all(
            [s == other.segments[i] for i, s in enumerate(self._segments)])

    def __hash__(self):
        return hash(tuple(hash(segment) for segment in self._segments))

    def __str__(self):
        return ".".join([str(s) for s in self.segments])

    def __add__(self, other):
        if isinstance(other, PackagePath):
            return PackagePath(self._segments + other.segments)
        elif isinstance(other, PackagePathSegment):
            copy = self._segments.copy()
            copy.append(other)
            return PackagePath(copy)
        elif isinstance(other, str):
            copy = self._segments.copy()
            copy.append(self.to_segments(other))
            return PackagePath(copy)


class SerializableMatcher:

    @staticmethod
    @abstractmethod
    def yaml_representer(dumper, data):
        """
        A yaml representer which can be used to dump a matcher to yaml
        :param dumper: yaml dumper
        :param data: data of the matcher to dump
        :return:
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def yaml_constructor(loader, node):
        """
        A yaml constructor to build a model from a yaml entry
        :param loader: yaml loader
        :param node: the node in the yaml
        :return:
        """
        raise NotImplementedError()

    @staticmethod
    def unserialize(reference):
        """
        Function to unserialize a matcher from string
        :param reference: String serialization
        :return: Matcher object
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize(self) -> str:
        """
        Function to simply serialize the matcher to string
        :return:
        """
        raise NotImplementedError()


class ISegmentMatcher(SerializableMatcher):
    """
    Matcher used to match a package path. This class is used to define mappings.
    """
    __metaclass__ = ABCMeta

    def __init__(self, content):
        """
        :param content: Content of the matcher. This can be everything used in the matches function.
        """
        self._content = content

    @property
    def content(self):
        """
        :return: Content of the matcher
        """
        return self._content

    @abstractmethod
    def matches(self, segment: PackagePathSegment):
        """
        Function used to check if a path segment matches the matcher
        :param segment:
        :return:
        """
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.content)

    def __eq__(self, other):
        return self.content == other.content

    def __str__(self):
        return self.serialize()


class StaticMatcher(ISegmentMatcher):
    """
    Matcher to directly match paths via string comparision
    """
    TAG = u'tag:yaml.org,2002:python/sSeg'

    @staticmethod
    def unserialize(reference) -> ISegmentMatcher:
        return StaticMatcher(reference)

    @staticmethod
    def yaml_representer(dumper, data):
        return dumper.represent_scalar(StaticMatcher.TAG, "%s" % data.conent)

    @staticmethod
    def yaml_constructor(loader, node):
        return StaticMatcher(loader.construct_scalar(node))

    def matches(self, segment: PackagePathSegment):
        return segment.content == self.content

    def serialize(self):
        return self.content


class RegexMatcher(ISegmentMatcher):
    """
    Matcher to use regex to match the given path segment.
    """
    TAG = u'tag:yaml.org,2002:python/rSeg'

    @staticmethod
    def unserialize(reference) -> ISegmentMatcher:
        return RegexMatcher(reference[4:-1])

    @staticmethod
    def yaml_representer(dumper, data):
        return dumper.represent_scalar(RegexMatcher.TAG, "%s" % data.conent)

    @staticmethod
    def yaml_constructor(loader, node):
        return RegexMatcher(loader.construct_scalar(node))

    def matches(self, segment: PackagePathSegment):
        return re.match(self.content, segment.content)

    def serialize(self):
        return "{re:" + self.content + "}"


class PackagePathMatcher(SerializableMatcher):
    """
    Matcher checking a complete path with multiple other matchers.
    """
    TAG = u'tag:yaml.org,2002:python/pPath'

    def __init__(self, content):
        if isinstance(content, str):
            content = self.to_matchers(content)
        self._matchers = content

    @property
    def matchers(self):
        """
        :return: SegmentMatchers of the PathMatcher
        """
        return self._matchers

    @staticmethod
    def unserialize(reference):
        return PackagePathMatcher(PackagePathMatcher.to_matchers(reference))

    @staticmethod
    def to_matchers(reference):
        """
        Convert a full reference string to multiple matchers "module.{re:foo}.bar" converts to static, regex, static
        :param reference:
        :return:
        """
        matchers = []
        for matcher_content in filter(lambda s: len(s) > 0, re.split(r"({.*?\})", reference)):
            if matcher_content.startswith("{re:"):
                matchers.append(RegexMatcher.unserialize(matcher_content))
            else:
                c: str
                matchers = matchers + [StaticMatcher.unserialize(c) for c in
                                       filter(lambda s: len(s) > 0, matcher_content.split("."))]
        return matchers

    @staticmethod
    def yaml_representer(dumper, data):
        return dumper.represent_scalar(PackagePathMatcher.TAG, "%s" % data.serialize())

    @staticmethod
    def yaml_constructor(loader, node):
        return PackagePathMatcher(loader.construct_scalar(node))

    def matches(self, package_path: PackagePath):
        return all([m.matches(package_path.segments[i]) for i, m in enumerate(self._matchers) if
                    i < len(package_path.segments)])

    def full_match(self, package_path: PackagePath):
        return len(self._matchers) == len(package_path.segments) and all(
            [m.matches(package_path.segments[i]) for i, m in enumerate(self._matchers)])

    def serialize(self):
        return ".".join([m.serialize() for m in self.matchers])

    def __add__(self, other):
        if isinstance(other, PackagePathMatcher):
            return PackagePathMatcher(self.matchers + other.matchers)
        elif isinstance(other, ISegmentMatcher):
            copy = self.matchers.copy()
            copy.append(other)
            return PackagePathMatcher(copy)
        elif isinstance(other, str):
            copy = self.matchers.copy()
            copy.append(self.to_matchers(other))
            return PackagePath(copy)


# Register yaml constructors for easy direct translation
yaml.add_constructor(StaticMatcher.TAG, StaticMatcher.yaml_constructor, Loader=yaml.SafeLoader)
yaml.add_constructor(RegexMatcher.TAG, RegexMatcher.yaml_constructor, Loader=yaml.SafeLoader)
yaml.add_constructor(PackagePathMatcher.TAG, PackagePathMatcher.yaml_constructor, Loader=yaml.SafeLoader)
yaml.add_representer(StaticMatcher, StaticMatcher.yaml_representer, Dumper=yaml.SafeDumper)
yaml.add_representer(RegexMatcher, RegexMatcher.yaml_representer, Dumper=yaml.SafeDumper)
yaml.add_representer(PackagePathMatcher, PackagePathMatcher.yaml_representer, Dumper=yaml.SafeDumper)



import os

from pydantic import HttpUrl, BaseModel
from typing import List, Type

from app.env import LoggerEnv
from pypads.app.injections.base_logger import TrackedObject
from pypads.app.injections.run_loggers import RunSetup
from pypads.app.misc.managed_git import ManagedGit
from pypads.model.models import TrackedObjectModel, TagMetaModel, ArtifactMetaModel, OutputModel
from pypads.utils.logging_util import WriteFormats


class GitTO(TrackedObject):
    class GitModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/SourceCode-Management"

        source: str = ...
        version: str = ...
        tags: List[TagMetaModel] = []
        git_log: ArtifactMetaModel = ...

        class Config:
            orm_mode = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.GitModel

    def __init__(self, *args, source, tracked_by, **kwargs):
        super().__init__(*args, source=source, tracked_by=tracked_by, **kwargs)

    def add_tag(self, name, value, description):
        meta = TagMetaModel(name=name, description=description)
        self.tags.append(meta)

        self._store_tag(value, meta)

    def store_git_log(self, name, value, format=WriteFormats.text):
        path = os.path.join(self._base_path(), self._get_artifact_path(name))
        self.git_log = ArtifactMetaModel(path=path, description="Commit logs for the git repository", format=format)

        self._store_artifact(value, self.git_log)

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), name)


class IGitRSF(RunSetup):
    """
    Function tracking the source code via git.
    """
    name = "Git Run Setup Logger"
    uri: HttpUrl = "https://www.padre-lab.eu/onto/git-run-logger"
    _dependencies = {"git"}

    class IGitRSFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/IGitRSF-Output"
        git_info: GitTO.get_model_cls() = None

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.IGitRSFOutput

    def _call(self, *args, _pypads_env: LoggerEnv, _logger_call, _logger_output, **kwargs):
        pads = _pypads_env.pypads
        _pypads_timeout = kwargs.get("_pypads_timeout") if kwargs.get("_pypads_timeout") else 5
        run = pads.api.active_run()
        tags = run.data.tags
        source_name = tags.get("mlflow.source.name", None)
        managed_git: ManagedGit = pads.managed_git_factory(source_name)
        if managed_git:
            repo = managed_git.repo
            git_info = GitTO(tracked_by=_logger_call, source=source_name or repo.working_dir, version=repo.head.commit.hexsha)
            # Disable pager for returns
            repo.git.set_persistent_git_options(no_pager=True)
            try:
                git_info.add_tag("pypads.git.description", repo.description, description="Repository description")
                git_info.add_tag("pypads.git.describe", repo.git.describe("--all"), description="")
                git_info.store_git_log("pypads.git.log", repo.git.log(kill_after_timeout=_pypads_timeout))
                remotes = repo.remotes
                remote_out = "No remotes existing"
                if len(remotes) > 0:
                    remote_out = ""
                    for remote in remotes:
                        remote_out += remote.name + ": " + remote.url + "\n"
                git_info.add_tag("pypads.git.remotes", remote_out, description="Remotes of the repositories")
            except Exception as e:
                _logger_output.set_failure_state(e)
            finally:
                git_info.store(_logger_output,"git_info")
